﻿#pragma once
#include "DataTyps.h"
#include "OSM.h"
#include "MAP.h"
#include "HeightData.h"
#include "Track.h"
#include "plot2D.h"
#include "fibunacci.h"
#include <QPainter>

typedef struct climbpoolele{
	float P;
	climb* c;
    Track* t;
    char* name;
}climbpoolele;
class nameaspirant{
    public:
        nameaspirant(char* name, float d);
        nameaspirant(char* name);
        float e;
        float s;
        float d;
        QString name;
        bool operator ==(nameaspirant o);
        bool operator < (const nameaspirant other) const{
            return d > other.d;
        }
};
//! extracts climbs from street network and provides climblayer handling
/*!
  A climb is defined by its minimum slope, is maximum descenting section, its
  minimum height meters and its minimum difficulty. It has to sufficiently
  independend from climbs which might lay at parts of the same Track.
  This class prides an algorithm to calculate all climbs. It is startet by
  calling the analyse() method. The algorithm works as follows:
  1.) put all cross into a height sorted list to serve as a climb start
  2.) lowest cross expands and labels the Neighbour cross, sets them active,
  i.g. put them into the list if they're not active
  3.) a cross will be reached by different climbs, some of them have to be
  stopped if too weak.
  4.) climb will only expand as long as the maximum descent fraction is not
  exceeded
  5.) finally the cross list will be empty
  6.) Now collect the climbs from the network beginning with the most difficult
  7.) only the most difficult and sufficently independend climbs shall survive
  The resulting climbpool can be used to generate a climblayer and html data
  etc.
  The climblayer is a layer on top of the MAP, displaying all climbs. This
  creates, saves und loads the climblayer.
*/
class ClimbAnalyse{
	public:
        ClimbAnalyse();                 /*!< use for tp load climblayer only*/
        //! starts a climb analysis
		ClimbAnalyse(OSM* O,short mintype, short maxtype, double minP,
                            double minSlope, double maxdwnHM, double minHM,
                            HeightData *h);
		~ClimbAnalyse();
        climblayer* get_ClimbLayer();   /*!< creates and returns climblayer*/
		climblayer* load_ClimbLayer(char* filename);
        void save_ClimbLayer(char* projectname, char* folder, climblayer *c);
		void create_MAP(char* projectname, char* folder, HeightData* h,
                        double size);
        void create_html(char* projectname, char* folder);
		void create_list(char* projectname, char* folder);
        //! restrict saved data to a rectangular area
        void set_brect(double minlat, double maxlat,
                       double minlon, double maxlon);
	private:
        void analyse();                 /*!< starts the actual analysis*/
        //! set the definitons of the climb
        void set_parameters(short mintype, short maxtype, double minP,
                            double minSlope, double maxdwnHM, double minHM);
        //!@{
        //!road type definition*/
        short mintype,maxtype;//!@}
        //!@{
        //!climb definition
        double minP, minSlope, maxdwnHM, minHM, autonomP;//!@}
        //! preparing cross to be inserted into f
		void reset_STRONG(KOO* k);
        //! priority queue to store active cross sorted by height
        FibunacciHeap<CAfib>* f;
        long int climbcount;            /*!< number of collected climbs*/
        //! container to store climbs when collect_climbs()
		climbpoolele* climbpool;
        void init_fib(KOO* k);          /*!< put all cross to f recursivly*/
        climb* cp;						/*!< climbpool, lists stopped climbs*/
		void init_climb(climb *c, KOO* k);
        //! makes this climb expand over all Neighbour relations
        /*! the used Neighbour relation has to fullfil the mintype and maxtype
            than the climb tries to expand. Whether the expansion is succesful
            or not is check by the storeif_climb() method, then the expansion
            is either accepted or refused.
        */
        void expand(climb* c);
        //! deletes a trackpointlist down to tpstop
		void free_tpl(trackpointlist* tp, trackpointlist* tpstop);
        //! adds an branch t to a axisting trackpointlist
		trackpointlist* add(trackpointlist* tp, trackpoint* t, int tpc,
							readdirection rd);
        //! calcs climb difficulty with fixed climb end
        float calc_P(trackpointlist* tp);
        //! calcs climb difficulty with loose climb end
        float calc_P(trackpoint* t,int arraylength);
        //! calcs climb difficulty with loose climb end
		float calc_Pmax(climb* c);
        //! calcs P, Pmax, Prec, tst of a climb
        void calc_climb(climb* c);
        //! determines, whether a new climb can stop an existing or vice versa
		short looses(climb* resident, climb* aspirant);
        //! checks whether the list represents a climb
		short is_climb(trackpointlist* tp);
        //! checks, whether used road are still avaiable in select_climbs()
        short all_active(climb* c);
        //! checks climb, whether used roads are sufficently independend
        short enough_active(climb* c);
        //! checks for existing climbs at cross and executes looses()
        short storeif_climb(Neighbour* n, trackpointlist* tp, climb* from);
        //! deletes the whole climb recursivly
        void del_climb(climb* c);
		void count_climbs(KOO* k);
        //! picks raw climbs from street net, to be selected (select_climbs())
		void collect_climbs(KOO* k);
        //! selects collected climbs, if suffiently independed or difficult
		void select_climbs();
        //! create names from OSM data
        void create_names();
        //! create climb name aspirant list according to street names
        void eval_Way_names(QList<nameaspirant> *wn, climb* c0);
        //! create climb name aspirant list from summits and other point data
        void eval_point_names(QList<nameaspirant> *pn, points **p, double fac,
                              Track* t, bool validstart);
        //! quicksort for collected climbs
		void sort_climbpool(long s, long e);
        //! trace back climb to start and creates Track instance from it
        void create_climb_Track(climbpoolele* c);
		short check(KOO* k);
        //! varifies, that climb has not used the cross at k before
		short crossunused(climb* c, KOO* k);
        OSM* o;                         /*!< holds the street network data */
        HeightData* h;                  /*!< for evaluating name aspirants */
        //! bounding rect, saved html data is limited to
        double minlat,maxlat,minlon,maxlon;
};
