#include "SummitRanking.h"

SummitRanking::SummitRanking(){}
SummitRanking::SummitRanking(QString worldfile, QString destfolder){
    FILE* f = fopen(worldfile.toLocal8Bit().data(),"r");
    if(f != NULL){
        char line[999];
        int tabs[7], itab,x,y,imax;
        double lon,lat;
        while(!feof(f)){
            fgets(line,999,f);
            itab = 0;
            imax = strlen(line);
            for(int i = 0; i < imax;i++)
                /*get fragments*/
                if(line[i] == '\t'){
                    tabs[itab] = i+1;
                    line[i] = '\0';
                    if(itab < 6)
                        itab++;
                    else
                        i = 1000;
                }
            if(line[tabs[5]] = 'T'){
                lon = atof(&(line[tabs[4]]));
                lat = atof(&(line[tabs[3]]));
                QString destfilename = QString("%3N%1E%2.txt").arg(floor(lat))
                                                            .arg(floor(lon))
                                                            .arg(destfolder);
                FILE* fdest = fopen(destfilename.toLocal8Bit().data(),"a");
                fprintf(fdest,"%f\t%f\t%s\n",lat,lon,&(line[tabs[0]]));
                fclose(fdest);
            }
        }
        fclose(f);
    }
}
SummitRanking::SummitRanking(int lat, int lon, QString srtm3folder,
                             QString destfolder){
    CalcSummits(lat, lon, srtm3folder, destfolder);
}
void SummitRanking::CalcSummits(int lat, int lon, QString srtm3folder,
                 QString destfolder){
    // check if file exists
    QString statusfile = QString(destfolder).append("status.txt");
    QString filename(srtm3folder);
    if(lat>0)
        filename.append(QString("N%1").arg(lat,2,10,QChar('0')));
    else
        filename.append(QString("S%1").arg(-lat,2,10,QChar('0')));
    if(lon>0)
        filename.append(QString("E%1.hgt").arg(lon,3,10,QChar('0')));
    else
        filename.append(QString("W%1.hgt").arg(-lon,3,10,QChar('0')));
    QFileInfo fi = QFileInfo(filename);
    if(fi.exists()){
        // test for high mountains
        HeightData htemp;
        htemp.set_SRTM3_folder(srtm3folder.toLocal8Bit().data());
        htemp.reset_t(lat,lat+1.0,lon,lon+1.0);
        if(htemp.max_dt()>100){
            //summits might be high enough
            prepare(srtm3folder.toLocal8Bit().data(),lon-1.5,lon+2.5,lat-1.5,lat+2.5);
            calc();
            summitlist* scounter = ssorted;
            cs = 0;
            while(scounter != NULL){
                // first enhance striking but not prominent summits
                if(scounter->data->parent != NULL){
                    if(scounter->data->Indexprominence > 0 &&
                       scounter->data->index > 0)
                        scounter->data->index = sqrt(scounter->data->index*scounter->data->Indexprominence);
                    else
                        scounter->data->index = 0;
                }
                // check if summits is within data area
                if(scounter->data->index > 0 &&
                   scounter->data->lat >= lat &&
                   scounter->data->lat < lat + 1 &&
                   scounter->data->lon >= lon &&
                   scounter->data->lon < lon + 1){
                    scounter->data->OldIndex = scounter->data->index;
                    scounter->data->index = 0;
                    //calc average fiets index
                    scounter->data->AvgIndex = 0;
                    for(int i = 0; i < 15; i++)
                        scounter->data->AvgIndex += scounter->data->AnglePmax[i];
                    scounter->data->AvgIndex = scounter->data->AvgIndex/15.;
                    //calc most impressive wall
                    scounter->data->Indexprominence = 0;
                    for(int i = 0; i < 15; i++)
                        scounter->data->Indexprominence +=
                             scounter->data->AnglePmax[i]
                           * scounter->data->AnglePmax[i];
                    scounter->data->Indexprominence = sqrt(scounter->data->Indexprominence/15.);
                    //calc variance
                    for(int i = 0; i < 15; i++)
                        scounter->data->index +=
                            (scounter->data->AnglePmax[i]-scounter->data->AvgIndex)
                           *(scounter->data->AnglePmax[i]-scounter->data->AvgIndex);
                    scounter->data->index = scounter->data->AvgIndex
                                          - sqrt(scounter->data->index/15.);
                }
                else
                    scounter->data->index = scounter->data->OldIndex = 0;
                scounter = scounter->next;
                cs++;
            }
            qsh = new summit*[cs];
            scounter = ssorted; cs = 0;
            while(scounter != NULL){
                qsh[cs]=scounter->data;
                scounter = scounter->next;
                cs++;
            }
            if(cs>0){
                sort_summit_list(0,cs-1);
                if(qsh[0]->AvgIndex > 10){
                    // assign names
                    QString destfilename = QString("%3N%1E%2.txt").arg(floor(lat))
                                                                .arg(floor(lon))
                                                                .arg(destfolder);
                    load_names(destfilename.toLocal8Bit().data());
                    // write data to hard disk
                    QString destlistname = QString("%3N%1E%2.lst").arg(floor(lat))
                                                                .arg(floor(lon))
                                                                .arg(destfolder);
                    FILE* fdest = fopen(destlistname.toLocal8Bit().data(),"w");
                    fprintf(fdest,"Index\tlar\tlon\theight\tprominence\tNewIndex\n");
                    for(long i = 0; i < cs; i++){
                        if(qsh[i]->AvgIndex > 10){
                            fprintf(fdest,"%f\t%f\t%f\t",qsh[i]->index,qsh[i]->lat,qsh[i]->lon);
                            fprintf(fdest,"%f\t%f\t%f\t%f\t",qsh[i]->height,
                                                             qsh[i]->prominence,
                                                             qsh[i]->OldIndex,
                                                             qsh[i]->Indexprominence);
                            summitname* snb = qsh[i]->names;
                            while(snb != NULL){
                                fprintf(fdest,"%s\t",snb->name);
                                snb = snb->next;
                            }
                            fprintf(fdest,"\n");
                        }
                    }
                    fclose(fdest);
                    emit log(statusfile,QString("%1\t%2\t%3\t%4 maximum peak\n")
                             .arg(QTime::currentTime().toString("hh:mm:ss"))
                             .arg(lat).arg(lon).arg(qsh[0]->index));
                }
                else{
                    emit log(statusfile,QString("%1\t%2\t%3\tno prominent peaks\n")
                             .arg(QTime::currentTime().toString("hh:mm:ss"))
                             .arg(lat).arg(lon));
                }
            }
            else{
                emit log(statusfile,QString("%1\t%2\t%3\tno peaks\n")
                         .arg(QTime::currentTime().toString("hh:mm:ss"))
                         .arg(lat).arg(lon));
            }
            delete qsh;
            unprepare();
        }
        else{
            emit log(statusfile,QString("%1\t%2\t%3\tno mountains\n")
                     .arg(QTime::currentTime().toString("hh:mm:ss"))
                     .arg(lat).arg(lon));
        }
    }
    else{
        emit log(statusfile,QString("%1\t%2\t%3\tno STRM3\n")
                 .arg(QTime::currentTime().toString("hh:mm:ss"))
                 .arg(lat).arg(lon));
    }
    this->~SummitRanking();
}

SummitRanking::SummitRanking(char* SRTM3path, char* summitlistfile,
				  double minlon, double maxlon, double minlat, double maxlat){
					  //shorten_list();
	prepare(SRTM3path,minlon,maxlon,minlat,maxlat);
	calc();
	/*sort summit list*/
	summitlist* scounter = ssorted;
	cs = 0;
	while(scounter != NULL){
		if(scounter->data->parent != NULL){
            if(scounter->data->Indexprominence > 0 && scounter->data->index > 0)
                scounter->data->index = sqrt(scounter->data->index*scounter->data->Indexprominence);
			else
				scounter->data->index = 0;
		}
        if(scounter->data->index > 0){
            scounter->data->OldIndex = 0;
            for(int i = 0; i < 15; i++)
                scounter->data->OldIndex += scounter->data->AnglePmax[i]
                                           *scounter->data->AnglePmax[i];
            scounter->data->OldIndex = sqrt(scounter->data->OldIndex)/15.;
            if(scounter->data->OldIndex > 10000)
                QMessageBox().exec();
            scounter->data->index = scounter->data->OldIndex;
        }
		scounter = scounter->next;
		cs++;
	}
	qsh = new summit*[cs];
	scounter = ssorted; cs = 0;
    while(scounter != NULL){
        qsh[cs]=scounter->data; scounter = scounter->next; cs++;}
	sort_summit_list(0,cs-1);
	load_names(summitlistfile);
	print();
}
SummitRanking::~SummitRanking(){
}
int SummitRanking::trace_uphill(double* lon, double* lat){
	/*follows gradient uphill to the next data cell and returns 1. Reaching a
	gradient zero returns 0. Both refresh stepped lat and lon. If trace runs
	out of data area, return 0 and set lon = lat = 1000;*/
	double stepsize = 0.0001, lc = 1. / cos(*lat/180. * 3.1415926);
	double dlon, dlat, scale, scalelimit, htemp;
	double dx = h.height_t(*lon,*lat,NULL,1)*lc;
	double dy = h.height_t(*lon,*lat,NULL,2);
	double hold = h.height_t(*lon,*lat,NULL,0);
	int xold = floor(*lon*1200);
	int yold = floor(*lat*1200);
    int TopReached = 0, OutOfData = 0, maxiter = 20;
	while( !TopReached && !OutOfData
		 && xold == floor(*lon*1200)
         && yold == floor(*lat*1200)
         && maxiter > 0){
		/*step small enough to reach higher point*/
        if(dx != 0 || dy != 0)
			scale = stepsize/sqrt(dx*dx+dy*dy);
		else
			{scale = 0;}
		scalelimit = 0.05 * scale;
		do{	dlon = dx * scale / lc;
			dlat = dy * scale;
			htemp = h.height_t(*lon+dlon,*lat+dlat,NULL,0);
			if(htemp == 0){OutOfData = 1; htemp = hold;}
			scale = scale * 0.5;
		}while(htemp < hold && scale > scalelimit);
		/*specify new point*/
		if(scale < scalelimit || (dx == 0 && dy == 0))
			TopReached = 1;
		hold = h.height_t(*lon+dlon,*lat+dlat,NULL,0);
		*lon = *lon + dlon;
		*lat = *lat + dlat;
		lc = 1. / cos(*lat/180. * 3.1415926);
		dx = h.height_t(*lon,*lat,NULL,1)*lc;
		dy = h.height_t(*lon,*lat,NULL,2);
        maxiter--;
	}
	if(OutOfData){
		*lon = *lat = 1000;
		return 0;
	}
    if(maxiter <= 0 || TopReached)
		return 0;
	else
		return 1;
}
summit* SummitRanking::create_summit(){
	summit* result = new summit;
	result->height = 0;
	result->index = 0;
	result->Indexprominence = 0;
	result->lat = 0;
	result->lon = 0;
	result->names = NULL;
	result->parent = NULL;
	result->prominence = NULL;
    result->names = NULL;
    for(int i = 0; i < 15; i++)
        result->AnglePmax[i] = 0;
	summitlist* sbuf = ssorted;
	ssorted = new summitlist;
	ssorted->data = result;
	ssorted->next = sbuf;
	return result;
}
void SummitRanking::sort_summit_list(long s, long e){
	summit* sb;
	long l = s, r = e;
	double p = qsh[(s+e)/2]->index;
	do{
		while(qsh[l]->index > p) l++;
		while(qsh[r]->index < p) r--;
		if(l<=r){
			sb = qsh[l]; qsh[l] = qsh[r]; qsh[r] = sb;
			l++;
			r--;
		}
	}while(l <= r);
	if(s < r) sort_summit_list(s,r);
	if(l < e) sort_summit_list(l,e);
}
double SummitRanking::hgp(summit* child){
	double result = -10000;
	while(child != NULL){
		result = child->height;
		child = child->parent;
	}
	return result;
}
void SummitRanking::inherit_parent(summit* taker, summit* giver,
								   double topoprom){
	summit* st = giver;
	while(giver->parent != NULL)
		giver = giver->parent;
	while(taker->parent != NULL)
		taker = taker->parent;
	taker->parent = giver;
	/*calc prominence if still required*/
	if(taker->Indexprominence == 0){
		taker->prominence = taker->height - topoprom;
		taker->Indexprominence = taker->index;
    }
}
void SummitRanking::prepare(char* SRTM3path,
				  double minlon, double maxlon, double minlat, double maxlat){
    ssorted = NULL;
	h.set_SRTM3_folder(SRTM3path);
    h.reset_t(minlat,maxlat,minlon,maxlon);
    GridRect = h.get_GridRect();
	cx = GridRect.maxlon-GridRect.minlon + 1;
	cy = GridRect.maxlat-GridRect.minlat + 1;
	s = new summit**[cy];
	for(int i = 0; i < cy; i++){
		s[i] = new summit*[cx];
        for(int j = 0; j < cx; j++)
			s[i][j] = NULL;
	}
    h.get_sorted_grid_data(&hs,&lons,&lats,&gradx,&grady,&count);
}
void SummitRanking::unprepare(){
    summitlist* sbuf;
    summitname* sb;
    while(ssorted != NULL){
        while(ssorted->data->names != NULL){
            sb = ssorted->data->names;
            ssorted->data->names = ssorted->data->names->next;
            if(sb->name != NULL)
                delete sb->name;
            delete sb;
        }
        delete ssorted->data;
        sbuf = ssorted;
        ssorted = ssorted->next;
        delete sbuf;
    }
    for(int i = 0; i < cy; i++){
        delete s[i];
    }
    delete s;
    delete grady;
    delete gradx;
    delete hs;
    delete lons;
    delete lats;
}
void SummitRanking::calc(){
    double d, lat0, lon0, dlon, dlat,h0,Acorr,dh;
    summit* st;
    //FILE* f = fopen("D:\\Summitranking\\trace.txt","w");
    for(long i = count-1; i >= 0; i--){
        int y0 = lats[i];
        int x0 = lons[i];
        lat0 = 1.*(GridRect.minlat+y0)/1200.;
        lon0 = 1.*(GridRect.minlon+x0)/1200.;
        Acorr = cos(lat0/180.*3.1415926);
        int y = grady[i];
        int x = gradx[i];
        if(y >= 0 && y < cy && x >= 0 && x < cx){
            /*summit is in that area*/
            if(s[y][x] == NULL){
                s[y][x] = create_summit();
                s[y][x]->lat = 1.*(GridRect.minlat+y)/1200.;
                s[y][x]->lon = 1.*(GridRect.minlon+x)/1200.;
                s[y][x]->height = h.height_t(s[y][x]->lon,s[y][x]->lat,NULL,0);

            }
            //fprintf(f,"%f\t%f\n\n",lons[i],lats[i]);
            //transfer summit
            s[y0][x0] = s[y][x];
            h0 = h.height_t(lon0,lat0,NULL,0);
            //check for inheriting parent summits
            for(int ix = -1; ix < 2; ix++)
                for(int iy = -1; iy < 2; iy++)
                    if(y0+iy >= 0 && y0+iy < cy && x0+ix >= 0 && x0+ix < cx &&
                      (ix != 0 || iy != 0))
                        if(s[y0+iy][x0+ix] != NULL){
                            if(hgp(s[y0+iy][x0+ix]) > hgp(s[y0][x0]))
                                inherit_parent(s[y0][x0],s[y0+iy][x0+ix],h0);
                        }
            //increase index of (all parent) summit
            st = s[y][x];
            while(st != NULL){
                dlon = st->lon - lon0;
                dlat = st->lat - lat0;
                dh = (st->height-h0);
                d = qMax(1.,
                         sqrt(dlat*dlat+dlon*dlon*Acorr*Acorr)*111194.926644559);
                double Ptemp = dh*dh/d;
                st->index += Ptemp/(d*d)*Acorr*1366.5564769531;
                // calc the old angular depend formula
                int angle_index = int((atan2(dlat,dlon*Acorr)+M_PI)
                                     /(2*M_PI)*15);
                if(angle_index >= 15)
                    angle_index = 0;
                if(angle_index < 0)
                    angle_index = 14;
                st->AnglePmax[angle_index] = qMax(Ptemp,st->AnglePmax[angle_index]);
                st = st->parent;
            }

        }
        else
            ;//fprintf(f,"\n");
    }
    //fclose(f);
}
void SummitRanking::load_names(char* summitlistfile){
	/*read geonames*/
	FILE* f = fopen(summitlistfile,"r");
	if(f != NULL){
		char line[999];
        int tabs[7], itab,x,y,imax;
        tabs[0] = 0;
		double lon,lat,lon0,lat0;
		while(!feof(f)){
			fgets(line,999,f);
            itab = 1;
			imax = strlen(line);
            line[imax-1] = '\0';
			for(int i = 0; i < imax;i++)
				/*get fragments*/
				if(line[i] == '\t'){
					tabs[itab] = i+1;
					line[i] = '\0';
					if(itab < 6)
						itab++;
					else
						i = 1000;
				}
            lon0 = lon = atof(&(line[tabs[1]]));
            lat0 = lat = atof(&(line[tabs[0]]));
            y = floor(lat*1200)-GridRect.minlat;
            x = floor(lon*1200)-GridRect.minlon;
            if(y >= 0 && y < cy && x >= 0 && x < cx){
                /*found summit within region*/
                while(trace_uphill(&lon,&lat)){}
                y = floor(lat*1200)-GridRect.minlat;
                x = floor(lon*1200)-GridRect.minlon;
                if(y >= 0 && y < cy && x >= 0 && x < cx
                   && s[y][x] != NULL){
                    /*summit point is within area*/
                    summitname* snbuf = s[y][x]->names;
                    s[y][x]->names = new summitname;
                    s[y][x]->names->lat = lat0;
                    s[y][x]->names->lon = lon0;
                    s[y][x]->names->height = h.height_t(lon0,lat0,NULL,0);
                    s[y][x]->names->distance = distance(lat0,lon0,lat,lon);
                    s[y][x]->names->next = snbuf;
                    s[y][x]->names->name = new char[strlen(&(line[tabs[2]]))+1];
                    strcpy(s[y][x]->names->name,&(line[tabs[2]]));
                }
            }
		}
		fclose(f);
	}
}
void SummitRanking::print(){
    FILE* f = fopen("D:\\Summitranking\\Heimat.lst","w");
	fprintf(f,"Index\tlar\tlon\theight\tprominence\tA\n");
	for(long i = 0; i < cs; i++){
		fprintf(f,"%f\t%f\t%f\t",qsh[i]->index,qsh[i]->lat,qsh[i]->lon);
		fprintf(f,"%f\t%f\t%f\t",qsh[i]->height,qsh[i]->prominence,qsh[i]->Indexprominence);
		summitname* snb = qsh[i]->names;
		while(snb != NULL){
			fprintf(f,"%s\t",snb->name);
			snb = snb->next;
		}
		fprintf(f,"\n");
	}
	fclose(f);
}
void SummitRanking::shorten_list(){
	FILE *f = fopen("C:\\Dokumente und Einstellungen\\Sirko\\Eigene Dateien\\Downloads\\allCountries.txt","r");
	FILE *f2 = fopen("C:\\Dokumente und Einstellungen\\Sirko\\Eigene Dateien\\Downloads\\LabelsHeimat.txt","w");
	if(f != NULL){
		char line[999];
		int tabs[7], itab,x,y,imax;
		double lon,lat,lon0,lat0;
		while(!feof(f)){
			fgets(line,999,f);
			itab = 0;
			imax = strlen(line);
			for(int i = 0; i < imax;i++)
				/*get fragments*/
				if(line[i] == '\t'){
					tabs[itab] = i+1;
					line[i] = '\0';
					if(itab < 6)
						itab++;
					else
						i = 1000;
				}
			if(line[tabs[5]] = 'T'){
				lon0 = lon = atof(&(line[tabs[4]]));
				lat0 = lat = atof(&(line[tabs[3]]));
				if(lon0 > 0 && lon0 < 20 && lat0 > 40 && lat0 < 55){
					/*found summit within region*/
					fprintf(f2,"%s\t",line);
					fprintf(f2,"%s\t",&(line[tabs[0]]));
					fprintf(f2,"%s\t",&(line[tabs[1]]));
					fprintf(f2,"%s\t",&(line[tabs[2]]));
					fprintf(f2,"%s\t",&(line[tabs[3]]));
					fprintf(f2,"%s\t",&(line[tabs[4]]));
					fprintf(f2,"%s\t\n",&(line[tabs[5]]));
				}
			}			
		}
		fclose(f);
		fclose(f2);
	}
}

