﻿#include <QtWidgets/QApplication>
#include <QFileInfo>
#include <QRegExp>
#include <QStringList>
#include "mainwindow.h"

int main(int argc, char *argv[])
{
    QApplication a(argc, argv);
    MainWindow w;
    w.show();
    //w.read_osm("D:\\ClimbProject\\pbf-Data\\problem5.osm");
    if(argc > 1){
        QRegExp rbound("(-b=)(-?\\d+\\.?\\d*)(,)(-?\\d+\\.?\\d*)(,)(-?\\d+\\.?\\d*)(,)(-?\\d+\\.?\\d*)");
        QRegExp rpngdest("(-m=)(\\d+)(,)(\\d+)(,)(.+)");
        QRegExp rbtpdest("(-d=)(.+)");
        QRegExp rosmread("(-r=)(.+)");
        QRegExp rbtpCA("(-a=)(.+)");
        double minlat = -90., maxlat = 90., minlon = -180., maxlon = 180.;
        int FalseCommand;
        for(int i = 1; i < argc; i++){
            FalseCommand = 1;
            if(rosmread.indexIn(argv[i])==0){
                // reading osm data
                w.read_osm(rosmread.cap(2));
                FalseCommand = 0;
            }
            if(rbtpdest.indexIn(argv[i])==0){
                // saving btp data within bounding rect
                w.save_btp(rbtpdest.cap(2),minlat,maxlat,minlon,maxlon);
                FalseCommand = 0;
            }
            if(rpngdest.indexIn(argv[i])==0){
                // generating and saving map
                MAP mt(w.hd,w.osm);
                mt.settings(&(w.c.MS));
                mt.resize(rpngdest.cap(2).toInt(),rpngdest.cap(4).toInt());
                mt.set_borders(minlat,maxlat,minlon,maxlon);
                mt.calc_borders();
                mt.draw_map(NULL,NULL,NULL,0);
                mt.save_to_png(rpngdest.cap(6));
                FalseCommand = 0;
            }
            if(rbtpCA.indexIn(argv[i])==0){
                ClimbAnalyse CA(w.osm,w.c.mintype,w.c.maxtype,w.c.CA.minP,
                                w.c.CA.minS,w.c.CA.maxdwnHM,w.c.CA.minHM,
                                w.hd);
                QString projectname = QDir(rbtpCA.cap(2)).dirName();
                CA.set_brect(minlat,maxlat,minlon,maxlon);
                CA.create_html(projectname.toLocal8Bit().data(),
                               rbtpCA.cap(2).toLocal8Bit().data());
                FalseCommand = 0;
            }
            if(rbound.indexIn(argv[i])==0){
                // setting up a bounding rect
                minlon = rbound.cap(2).toDouble();
                minlat = rbound.cap(4).toDouble();
                maxlon = rbound.cap(6).toDouble();
                maxlat = rbound.cap(8).toDouble();
                FalseCommand = 0;
            }
            if(QString(argv[i])=="close"){
                w.autosave = 0;
                return w.close();
            }
            if(FalseCommand){
                if(QFileInfo(QString(argv[i])).suffix()=="btp"){
                    // loading data set
                    w.load_btp(QString().fromLocal8Bit(argv[i]));
                    FalseCommand = 0;
                }
                if(QFileInfo(QString(argv[i])).suffix()=="gpx"){
                    // loading and analysing gpx file
                    w.analyse_gpx(QString().fromLocal8Bit(argv[i]));
                    FalseCommand = 0;
                }
                if(QFileInfo(QString(argv[i])).suffix()=="osm"){
                    // reading osm data
                    w.read_osm(QString().fromLocal8Bit(argv[i]));
                    FalseCommand = 0;
                }
                if(FalseCommand)
                    // still not recognized commend
                    QMessageBox(QMessageBox::NoIcon,"Unknown argument",
                                QString().fromLocal8Bit(argv[i])).exec();
            }
        }
    }
    
    return a.exec();
}

