﻿#include "osm.h"

OSM::OSM(string fileName){
	setfilename(fileName);
	initiallists();
}
OSM::OSM(){
	initiallists();
}
OSM::~OSM(){
    delete_AVL(*(D.nodes));
    delete_AVL(*(D.waysavl));
    delete_lines(*motorWay);
    delete_lines(*highway);
    delete_lines(*primary);
    delete_lines(*secondary);
    delete_lines(*track);
    delete_lines(*track2);
    delete_lines(*creek);
    delete_lines(*river);
    delete_lines(*NoHeight);
    delete_lines(*wood);
    delete_lines(* water);
    delete_points(*metropolis);
    delete_points(*city);
    delete_points(*town);
    delete_points(*village);
    delete_points(*locality);
    delete_points(*col);
    delete_points(*peak);
    delete_relation(*woodr);
    delete_relation(*waterr);
}
qint64 OSM::extract_valuei(char* s, const char* a){
	char *x1;
	char *x2;
	x1 = strstr(s,a)+strlen(a)+2*sizeof(char);
    if(x1 == NULL)
        QMessageBox(QMessageBox::NoIcon,"Error",QString("Could not extract integer value from '%1'").arg(s)).exec();
	x2 = strchr(x1,'\'');
	*x2 = '\0';
    qint64 result = QString(x1).toLongLong();
	*x2 = '\'';
	return result;
}
double OSM::extract_valuef(char* s, const char* a){
	char *x1;
	char *x2;
	x1 = strstr(s,a)+strlen(a)+2*sizeof(char);
    if(x1 == NULL)
        QMessageBox(QMessageBox::NoIcon,"Error",QString("Could not extract double value from '%1'").arg(s)).exec();

	x2 = strchr(x1,'\'');
	*x2 = '\0';
	double result = atof(x1);
	*x2 = '\'';
	return result;
}
void OSM::read_OSM(){
	if(status >= 1){
		FILE *f;
		f = fopen(filename.c_str(),"r");
		if (f == NULL){
			printf("Fehler beim öffnen der Datei ");
			printf(filename.c_str());
		}
		else{
            char line[10000],lineattr[10000],linename[10000];

            qint64 currentseek;
			dminlat = 0;
            dmaxlat = 0;
			dminlon = 0;
			dmaxlon = 0;

            fseek(f,0,SEEK_SET);
			while(!feof(f)){
                currentseek = _ftelli64(f);
				fgets(line,999,f);
				quotationreplace(line);
                if(strstr(line,"bbox=") != NULL
					|| strstr(line,"<bound box=") != NULL)
					extract_bbox(line);
                if(strstr(line,"<bounds ") != NULL){
                    dminlat = extract_valuef(line," minlat");
                    dmaxlat = extract_valuef(line," maxlat");
                    dminlon = extract_valuef(line," minlon");
                    dmaxlon = extract_valuef(line," maxlon");
				}
                if (strstr(line,"<node ") != NULL)
					nodehandling(f,line,lineattr,linename);
                if (strstr(line,"<way ") != NULL)
                    Wayhandling(f,line,lineattr,linename,&currentseek);
                if (strstr(line,"<relation ") != NULL)
                    relationhandling(f,line,lineattr,linename,&currentseek);
			}
			fclose(f);
			status = 2;
		}
	}
	else{
		printf("Initialisieren korrekten Dateinamen mit OSM(fileName)!\n");
	}
}
KOO* OSM::search(qint64 id, KOO* s){
	while (s != NULL && s->id != id){
		if(s->id > id) s = s->l;
		else s = s->r;
	}
	return s;
}
KOO* OSM::insert(qint64 id, KOO* s){
	KOO* p = s;
	if (s != NULL){
		/*Falls der Baum schon angelegt ist*/
		while (s != NULL && s->id != id){
			/*suche das zugehörige Blatt*/
			if(s->id > id){
				p = s;
				s = s->l;
			}
			else
				if(s->id < id){
					p = s;
					s = s->r;
				}
		}
		if(s != NULL)
			/*das Blatt ist schon vorhanden*/
			return s;
		else{
			/*Blatt muss angelegt werden*/
			if(p->id > id){
				/*Blatt links*/
                p->l = new KOO;
				s = p->l;
				s->prev = p;
				s->id = id;
				s->l = NULL;
				s->r = NULL;
				s->lat = 200;
				s->lon = 200;
				s->balance = 0;
				s->status = -1;
				s->cd = NULL;
				return s;
			}
			else{
				/*Blatt rechts*/
                p->r = new KOO;
				s = p->r;
				s->prev = p;
				s->id = id;
				s->l = NULL;
				s->r = NULL;
				s->lat = 200;
				s->lon = 200;
				s->balance = 0;
				s->status = -1;
				s->cd = NULL;
				return s;
			}
		}
	}
	else{
		/*Baum anlegen, weil noch nicht geschehen*/
        s = new KOO;
		s->prev = NULL;
		s->id = id;
		s->l = NULL;
		s->r = NULL;
		s->lat = 200;
		s->lon = 200;
		s->balance = 0;
		s->status = -1;
		s->cd = NULL;
		return s;
	}
}
KOO* OSM::rotate(KOO** s, branch b){
	KOO* k;
	switch(b){
		case l:	switch((*s)->balance){
				case +1:switch((*s)->r->balance){
						case -1:(*s)->balance = 0;
								(*s)->r->balance = -2;
								break;
						case 0: (*s)->balance = 0;
								(*s)->r->balance = -1;
								break;
						case 1: (*s)->balance = -1;
								(*s)->r->balance = -1;
								break;
						}
						break;
				case +2:switch((*s)->r->balance){
						case -1:(*s)->balance = 1;
								(*s)->r->balance = -2;
								break;
						case 0: (*s)->balance = 1;
								(*s)->r->balance = -1;
								break;
						case 1: (*s)->balance = 0;
								(*s)->r->balance = 0;
								break;
						case 2: (*s)->balance = -1;
								(*s)->r->balance = 0;
								break;
						}
						break;
				}
				break;
		case r:	switch((*s)->balance){
				case -1:switch((*s)->l->balance){
						case -1:(*s)->balance = 1;
								(*s)->l->balance = 1;
								break;
						case 0: (*s)->balance = 0;
								(*s)->l->balance = 1;
								break;
						case 1: (*s)->balance = 0;
								(*s)->l->balance = 2;
								break;
						}
						break;
				case -2:switch((*s)->l->balance){
						case -2:(*s)->balance = 1;
								(*s)->l->balance = 0;
								break;
						case -1:(*s)->balance = 0;
								(*s)->l->balance = 0;
								break;
						case 0: (*s)->balance = -1;
								(*s)->l->balance = 1;
								break;
						case 1: (*s)->balance = -1;
								(*s)->l->balance = 2;
								break;
						}
						break;
				}
				break;
    default: break;
	}
	if (b == l){
		k = (*s)->r->l;
		(*s)->r->l = (*s);
		(*s) = (*s)->r;
		(*s)->l->r = k;
		(*s)->prev = (*s)->l->prev;
		(*s)->l->prev = (*s);
		if (k != NULL) k->prev = (*s)->l;
		return (*s);
	}
	else{
		k = (*s)->l->r;
		(*s)->l->r = (*s);
		(*s) = (*s)->l;
		(*s)->r->l = k;
		(*s)->prev = (*s)->r->prev;
		(*s)->r->prev = (*s);
		if (k != NULL) k->prev = (*s)->r;
		return (*s);
	}
}
KOO* OSM::insert_AVL(qint64 id, KOO** s){
	if(*s != NULL){
		int fertig = 0;
		KOO* i = insert(id,*s);
		KOO* result = i;
		if(i->status > -1)				//Falls die KOO schon vorhanden war 
			return i;					//ist die gesammte Routine zu beenden

		i->status = 0;
		while(!fertig){
			i = i->prev;
			if (id < i->id){
				i->balance--;
				switch(i->balance){
					case-2: switch(i->l->balance){
								case -1: i = rotate(&i,r);
										 break;
								case  1: rotate(&(i->l),l);
										 i = rotate(&i,r);
										 break;
							}
							fertig = 1;
							if(i->prev != NULL){
								if(i->prev->id > i->id) 
									i->prev->l = i;
								else
									i->prev->r = i;
							}
							break;
					case-1: if(i->prev == NULL) fertig = 1;
							break;
					case 0: fertig = 1;
							break;
				}
			}
			else{
				i->balance++;
				switch(i->balance){
					case 2: switch(i->r->balance){
								case  1: i = rotate(&i,l);
										 break;
								case -1: rotate(&(i->r),r);
										 i = rotate(&i,l);
										 break;
							}
							fertig = 1;
							if(i->prev != NULL){
								if(i->prev->id > i->id) 
									i->prev->l = i;
								else
									i->prev->r = i;
							}
							break;
					case 1: if(i->prev == NULL) fertig = 1;
							break;
					case 0: fertig = 1;
							break;
				}
			}
		}
		if (i->prev == NULL){
			*s = i;
			return result;
		}
		else 
			return result;
	}
	else{
    KOO* result = new KOO;
	result->balance = 0;
	result->id = id;
	result->prev = NULL;
	result->l = NULL;
	result->r = NULL;
	result->lat = 200;
	result->lon = 200;
	result->status = 0;
	result->cd = NULL;
	*s = result;
	return result;
	}
}
Way* OSM::search(qint64 id, Way* s){
	while (s != NULL && s->id != id){
		if(s->id > id) s = s->l;
		else s = s->r;
	}
	return s;
}
Way* OSM::insert(qint64 id, Way* s){
    Way* p = s;
	if (s != NULL){
		/*Falls der Baum schon angelegt ist*/
		while (s != NULL && s->id != id){
			/*suche das zugehörige Blatt*/
			if(s->id > id){
				p = s;
				s = s->l;
			}
			else
				if(s->id < id){
					p = s;
					s = s->r;
				}
		}
		if(s != NULL)
			/*das Blatt ist schon vorhanden*/
			return s;
		else{
			/*Blatt muss angelegt werden*/
			if(p->id > id){
				/*Blatt links*/
                p->l = new Way;
				s = p->l;
				s->prev = p;
				s->id = id;
				s->l = NULL;
				s->r = NULL;
				s->balance = 0;
				s->type = -1;
				return s;
			}
			else{
				/*Blatt rechts*/
                p->r = new Way;
				s = p->r;
				s->prev = p;
				s->id = id;
				s->l = NULL;
				s->r = NULL;
				s->balance = 0;
				s->type = -1;
				return s;
			}
		}
	}
	else{
		/*Baum anlegen, weil noch nicht geschehen*/
        s = new Way;
		s->prev = NULL;
		s->id = id;
		s->l = NULL;
		s->r = NULL;
		s->balance = 0;
		s->type = -1;
		return s;
	}
}
Way* OSM::rotate(Way** s, branch b){
    Way* k;
	switch(b){
		case l:	switch((*s)->balance){
				case +1:switch((*s)->r->balance){
						case -1:(*s)->balance = 0;
								(*s)->r->balance = -2;
								break;
						case 0: (*s)->balance = 0;
								(*s)->r->balance = -1;
								break;
						case 1: (*s)->balance = -1;
								(*s)->r->balance = -1;
								break;
						}
						break;
				case +2:switch((*s)->r->balance){
						case -1:(*s)->balance = 1;
								(*s)->r->balance = -2;
								break;
						case 0: (*s)->balance = 1;
								(*s)->r->balance = -1;
								break;
						case 1: (*s)->balance = 0;
								(*s)->r->balance = 0;
								break;
						case 2: (*s)->balance = -1;
								(*s)->r->balance = 0;
								break;
						}
						break;
				}
				break;
		case r:	switch((*s)->balance){
				case -1:switch((*s)->l->balance){
						case -1:(*s)->balance = 1;
								(*s)->l->balance = 1;
								break;
						case 0: (*s)->balance = 0;
								(*s)->l->balance = 1;
								break;
						case 1: (*s)->balance = 0;
								(*s)->l->balance = 2;
								break;
						}
						break;
				case -2:switch((*s)->l->balance){
						case -2:(*s)->balance = 1;
								(*s)->l->balance = 0;
								break;
						case -1:(*s)->balance = 0;
								(*s)->l->balance = 0;
								break;
						case 0: (*s)->balance = -1;
								(*s)->l->balance = 1;
								break;
						case 1: (*s)->balance = -1;
								(*s)->l->balance = 2;
								break;
						}
						break;
				}
				break;
    default: break;
	}
	if (b == l){
		k = (*s)->r->l;
		(*s)->r->l = (*s);
		(*s) = (*s)->r;
		(*s)->l->r = k;
		(*s)->prev = (*s)->l->prev;
		(*s)->l->prev = (*s);
		if (k != NULL) k->prev = (*s)->l;
		return (*s);
	}
	else{
		k = (*s)->l->r;
		(*s)->l->r = (*s);
		(*s) = (*s)->l;
		(*s)->r->l = k;
		(*s)->prev = (*s)->r->prev;
		(*s)->r->prev = (*s);
		if (k != NULL) k->prev = (*s)->r;
		return (*s);
	}
}
Way* OSM::insert_AVL(qint64 id, Way** s){
	if(*s != NULL){
		int fertig = 0;
        Way* i = insert(id,*s);
		if(i->type > -1)				//der Weg war schon vorhanden und die 
			fertig = 1;					//Routine kann beendet werden
        Way* result = i;
		while(!fertig){
			i = i->prev;
			if (id < i->id){
				i->balance--;
				switch(i->balance){
					case-2: switch(i->l->balance){
								case -1: i = rotate(&i,r);
										 break;
								case  1: rotate(&(i->l),l);
										 i = rotate(&i,r);
										 break;
							}
							fertig = 1;
							if(i->prev != NULL){
								if(i->prev->id > i->id) 
									i->prev->l = i;
								else
									i->prev->r = i;
							}
							break;
					case-1: if(i->prev == NULL) fertig = 1;
							break;
					case 0: fertig = 1;
							break;
				}
			}
			else{
				i->balance++;
				switch(i->balance){
					case 2: switch(i->r->balance){
								case  1: i = rotate(&i,l);
										 break;
								case -1: rotate(&(i->r),r);
										 i = rotate(&i,l);
										 break;
							}
							fertig = 1;
							if(i->prev != NULL){
								if(i->prev->id > i->id) 
									i->prev->l = i;
								else
									i->prev->r = i;
							}
							break;
					case 1: if(i->prev == NULL) fertig = 1;
							break;
					case 0: fertig = 1;
							break;
				}
			}
		}
		if (i->prev == NULL){
			*s = i;
			return result;
		}
		else 
			return result;
	}
	else{
    Way* result = new Way;
	result->balance = 0;
	result->id = id;
	result->prev = NULL;
	result->l = NULL;
	result->r = NULL;
    result->nodec = 0;
	result->type = -1;
	*s = result;
	return result;
	}
}
void OSM::get_KOO_Waysavl(){
    get_KOO_Waysavl(*(D.waysavl),1);
}
void OSM::get_KOO_Waysavl(Way* l,short next){
	if(status >=2){
		status = 3;
        qint64 id;
		if(l != NULL){
            l->maxlat = -90;
			l->maxlon = -180;
			l->minlat =  90;
			l->minlon =  180;

            for(int j = 0; j < l->nodec; j++){
				id = l->node[j]->id;
                delete(l->node[j]);
				l->node[j] = search(id, *D.nodes);
				if(l->node[j] == NULL){
					l->type = 99;
					break;
				}

				/*Bestimme das umrandende Rechteck*/
				if(l->node[j]->lat > l->maxlat ) 
					l->maxlat = float(l->node[j]->lat);
				if(l->node[j]->lat < l->minlat ) 
					l->minlat = float(l->node[j]->lat);
				if(l->node[j]->lon > l->maxlon ) 
					l->maxlon = float(l->node[j]->lon);
				if(l->node[j]->lon < l->minlon ) 
					l->minlon = float(l->node[j]->lon);
			}
			if(next){
                get_KOO_Waysavl(l->l,1);	//rekursiver Aufruf des linken Kindes
                get_KOO_Waysavl(l->r,1);	//rekursiver Aufruf des rechten Kindes
			}
		}
	}
	else{
		printf("Zuerst Datei einlesen mittels read_OSM_all()\n");
	}
}
char* OSM::extract_valuec(char* s, const char* a){
	char *x1;
	char *x2;
	x1 = strstr(s,a)+strlen(a)+2*sizeof(char);
    if(x1 == NULL)
        QMessageBox(QMessageBox::NoIcon,"Error",QString("Could not char value from '%1'").arg(s)).exec();
	x2 = strchr(x1,'\'');
	*x2 = '\0';
	return x1;	
}
void OSM::extract_bbox(char* s){
	if(strstr(s,"bbox=") != NULL){
		char* t = strstr(s,"bbox=");
		t = &(t[5]);
		char* t2 = strstr(t,",");
		t2[0]='\0';
		dminlon = atof(t);
		t = &(t2[1]);
		t2 = strstr(t,",");
		t2[0]='\0';
		dminlat = atof(t);
		t = &(t2[1]);
		t2 = strstr(t,",");
		t2[0]='\0';
		dmaxlon = atof(t);
		t = &(t2[1]);
		t2 = strstr(t,"'");
		t2[0]='\0';
		dmaxlat = atof(t);
	}
	else{
		char* t = strstr(s,"box=");
		t = &(t[5]);
		char* t2 = strstr(t,",");
		t2[0]='\0';
		dminlat = atof(t);
		t = &(t2[1]);
		t2 = strstr(t,",");
		t2[0]='\0';
		dminlon = atof(t);
		t = &(t2[1]);
		t2 = strstr(t,",");
		t2[0]='\0';
		dmaxlat = atof(t);
		t = &(t2[1]);
		t2 = strstr(t,"'");
		t2[0]='\0';
		dmaxlon = atof(t);
	}
}
short OSM::isin(lines* l,double minlat, double maxlat, 
						 double minlon, double maxlon){
    if	  ( ( (l->way->minlat >= minlat && l->way->minlat <= maxlat)
            ||(l->way->maxlat <= maxlat && l->way->maxlat >= minlat)
            ||(l->way->minlat <= minlat && l->way->maxlat >= maxlat)
			)
		&&
            ( (l->way->minlon >= minlon && l->way->minlon <= maxlon)
            ||(l->way->maxlon <= maxlon && l->way->maxlon >= minlon)
            ||(l->way->minlon <= minlon && l->way->maxlon >= maxlon)
			) )
		return 1;
	else
		return 0;
}
short OSM::isin(points* l,double minlat, double maxlat, 
						  double minlon, double maxlon){
    return isin(l->k,minlat,maxlat,minlon,maxlon);
}
short OSM::isin(KOO* k,double minlat, double maxlat, 
					   double minlon, double maxlon){
	if(    k->lat < maxlat
		&& k->lat > minlat
		&& k->lon > minlon
		&& k->lon < maxlon)
		return 1;
	else
		return 0;
}
short OSM::isin(Way* w,double minlat, double maxlat,
					   double minlon, double maxlon){
    if	  ( ( (w->minlat >= minlat && w->minlat <= maxlat)
            ||(w->maxlat <= maxlat && w->maxlat >= minlat)
            ||(w->minlat <= minlat && w->maxlat >= maxlat)
			)
		&&
            ( (w->minlon >= minlon && w->minlon <= maxlon)
            ||(w->maxlon <= maxlon && w->maxlon >= minlon)
            ||(w->minlon <= minlon && w->maxlon >= maxlon)
			) )
		return 1;
	else
		return 0;
}
short OSM::iscompletein(Way* w,double minlat, double maxlat,
					   double minlon, double maxlon){
	if(	   w->minlat > minlat
		&& w->maxlat < maxlat
		&& w->minlon > minlon
		&& w->maxlon < maxlon)
		return 1;
	else
		return 0;
}
void OSM::setfilename(string fileName){
	if(fileName.compare(fileName.size()-4,4,".osm") == 0){
		filename = fileName;
		status = 1;
	}
	else{
		printf("%s ist keien .osm-Datei!\n",fileName.c_str());
		status = 0;
	}
}
void OSM::initiallists(){
	maxlat = -90;
	minlat = 90;
	maxlon = -180;
	minlon = 180;
	dmaxlat = -90;
	dminlat = 90;
	dmaxlon = -180;
	dminlon = 180;
	
    D.nodes = new KOO*;
    D.waysavl = new Way*;
	*(D.nodes) = NULL;
    *(D.waysavl) = NULL;
	
    motorWay = new lines*;
    highway = new lines*;
    primary = new lines*;
    secondary = new lines*;
    track = new lines*;
    track2 = new lines*;
    creek = new lines*;
    river = new lines*;
    NoHeight = new lines*;
    *motorWay = *highway = *primary = *secondary = *track = *track2
	= *creek = *river = *NoHeight = NULL;
	
    wood =new lines*;
    water =new lines*;
	*wood = *water = NULL;

    metropolis = new points*;
    city = new points*;
    town = new points*;
    village = new points*;
    locality = new points*;
    col = new points*;
    *metropolis = *city = *town = *village = *locality = *col = NULL;

    woodr = new relation*;
    waterr = new relation*;
	*woodr = *waterr = NULL;

    peak = new points*;
	*peak = NULL;

    ExclCheck[0] = motorWay;
    ExclCheck[1] = highway;
	ExclCheck[2] = primary;
	ExclCheck[3] = secondary;
	ExclCheck[4] = track;
	ExclCheck[5] = track2;
	ExclCheck[6] = river;
	ExclCheck[7] = creek;

}
int  OSM::get_status(){
	return status;
}
int OSM::count_cross(KOO *k){
    if(k != NULL){
        if(k->cd != NULL)
            return 1+count_cross(k->l)+count_cross(k->r);
        else
            return count_cross(k->l)+count_cross(k->r);
    }
    else
        return 0;
}
int OSM::count_STRONG(KOO *k){
    if(k != NULL){
        if(k->cd != NULL && k->cd->s != NULL)
            return 1+count_STRONG(k->l)+count_STRONG(k->r);
        else
            return count_STRONG(k->l)+count_STRONG(k->r);
    }
    else
        return 0;
}

void OSM::nodehandling(FILE *f,char* line,char* lineattr,char* linename){
	char* attr = &line[0];
	char* name = NULL;
    short foundplace = 0,foundpeak = 0, foundpopulation = 0, foundname = 0, foundcol = 0;
	int population = 0;
	int normalname = 0;
	 
	
	KOO* k;
	k = insert_AVL(extract_valuei(line," id"),D.nodes);
    k->lat = extract_valuef(line," lat");
    k->lon = extract_valuef(line," lon");
	k->cd = NULL;
	
    while(strstr(line,"</node>") == NULL &&
          !(strstr(line,"<node ")!=NULL && strstr(line,"/>") != NULL)){
				fgets(line,999,f);
				quotationreplace(line);
                if(strstr(line,"<tag k='place' v='") != NULL){
					strcpy(lineattr,line);
					attr = extract_valuec(lineattr,"v");
					foundplace = 1;
				}
                if(strstr(line,"<tag k='tourism' v='alpine_hut'/>") != NULL){
                    attr = "alpine_hut";
                    foundplace = 1;
                }
                if(strstr(line,"<tag k='natural' v='peak'/>") != NULL){
                    foundpeak = 1;
                }
                if(strstr(line,"<tag k='natural' v='saddle'/>") != NULL ||
                   strstr(line,"<tag k='mountain_pass' v='yes'/>") != NULL){
                    foundcol = 1;
				}
                if(strstr(line,"<tag k='name' v='") != NULL){
                    // primary name source
					strcpy(linename,line);
					name = extract_valuec(linename,"v");
                    normalname = 2;
                    foundname = 1;
				}
                if(normalname < 2 &&
                   strstr(line,"<tag k='openGeoDB:name' v='")!=NULL){
                    // secondary name source
                    strcpy(linename,line);
                    name = extract_valuec(linename,"v");
                    normalname = 1;
                    foundname = 1;
                }
                if(normalname < 1 &&
                    strstr(line,"<tag k='name:")!=NULL){
                    // tertiary name source
					strcpy(linename,line);
					name = extract_valuec(linename,"v");
					foundname = 1;
				}
                if(strstr(line,"tag k='population' v='") != NULL){
					population = extract_valuei(line,"v");
					foundpopulation = 1;
				}
	}
    //falls kein Name gefunden sollte Punkt nicht verzeichnet werden
	if(name == NULL)
		foundplace = foundpeak = 0;
	else
		htmlreplace(name);

    if(foundpopulation && foundplace && !foundcol)
        //d.h. ein Ort mit Einwohnerzahlangabe ist ausgelesen wurden
        if(population > 0){
            //Klassifizierung nach Einwohnerzahl ist sicherer
			if(population>100000)
				add_points(metropolis,k,name,float(population));
            else{
				if(population>10000)
                    add_points(city,k,name,float(population));
                else{
					if(population>3000 || !strcmp(attr,"town"))
					add_points(town,k,name,float(population));
					else
                        if(population>50)
                            add_points(village,k,name,float(population));
                        else
                            add_points(locality,k,name,float(population));

                }
            }
        }
    if(foundplace && !foundpopulation  && !foundcol){
            //...als Klassifizierung nach OSM-tags, da keine Einwohnerzahl
			if(!strcmp(attr,"city"))
				add_points(metropolis,k,name,500000);
            else{
				if(!strcmp(attr,"town"))
					add_points(town,k,name,5000);
                else{
                    if(   !strcmp(attr,"suburb")
                        ||!strcmp(attr,"village"))
						add_points(village,k,name,500);
                    else
                        if(  !strcmp(attr,"hamlet")
                           ||!strcmp(attr,"locality")
                           ||!strcmp(attr,"alpine_hut")
                           ||!strcmp(attr,"isolated_dwelling"))
                            add_points(locality,k,name,20);
                }
            }
    }
    if(foundcol && foundname){
        add_points(col,k,name,20);
    }
    if(foundpeak && foundname && !foundcol)
		add_points(peak,k,name,0);
}

/*void OSM::nodehandling(FILE *f,char* line,char* lineattr,char* linename){
    //SKIKEmap
    char* attr = &line[0];
    char* name = NULL;
    short foundplace = 0,foundpeak = 0, foundpopulation = 0, foundname = 0, foundcol = 0;
    int population = 0;
    int normalname = 0;


    KOO* k;
    k = insert_AVL(extract_valuei(line," id"),D.nodes);
    k->lat = extract_valuef(line," lat");
    k->lon = extract_valuef(line," lon");
    k->cd = NULL;

    while(strstr(line,"</node>") == NULL &&
          !(strstr(line,"<node ")!=NULL && strstr(line,"/>") != NULL)){
                fgets(line,999,f);
                quotationreplace(line);
                if(strstr(line,"<tag k='place' v='") != NULL){
                    strcpy(lineattr,line);
                    attr = extract_valuec(lineattr,"v");
                    foundplace = 1;
                }
                if(strstr(line,"<tag k='tourism' v='alpine_hut'/>") != NULL){
                    attr = "alpine_hut";
                    foundplace = 1;
                }
                if( strstr(line,"<tag k='shop' v='supermarket'/>") != NULL||
                    strstr(line,"<tag k='shop' v='general'/>") != NULL||
                    strstr(line,"<tag k='shop' v='convenience'/>") != NULL||
                    strstr(line,"<tag k='amenity' v='bar'/>") != NULL||
                    strstr(line,"<tag k='amenity' v='bar'/>") != NULL||
                    strstr(line,"<tag k='amenity' v='bbq'/>") != NULL||
                    strstr(line,"<tag k='amenity' v='biergarten'/>") != NULL||
                    strstr(line,"<tag k='amenity' v='cafe'/>") != NULL||
                    strstr(line,"<tag k='amenity' v='fast_food'/>") != NULL||
                    strstr(line,"<tag k='amenity' v='food_court'/>") != NULL||
                    strstr(line,"<tag k='amenity' v='ice_cream'/>") != NULL||
                    strstr(line,"<tag k='amenity' v='restaurant'/>") != NULL||
                    strstr(line,"<tag k='amenity' v='fuel'/>") != NULL||
                    strstr(line,"<tag k='amenity' v='pub'/>") != NULL){
                    attr = "city";
                    foundplace = 1;
                }
                if(strstr(line,"<tag k='natural' v='peak'/>") != NULL){
                    foundpeak = 1;
                }
                if(strstr(line,"<tag k='natural' v='saddle'/>") != NULL ||
                   strstr(line,"<tag k='mountain_pass' v='yes'/>") != NULL||
                   strstr(line,"<tag k='railway' v='station'/>") != NULL||
                   strstr(line,"<tag k='railway' v='halt'/>") != NULL||
                   strstr(line,"<tag k='amenity' v='bus_station'/>") != NULL||
                   strstr(line,"<tag k='public_transport' v='station'/>") != NULL){
                   foundcol = 1;
                }
                if(strstr(line,"<tag k='name' v='") != NULL){
                    // primary name source
                    strcpy(linename,line);
                    name = extract_valuec(linename,"v");
                    normalname = 2;
                    foundname = 1;
                }
                if(normalname < 2 &&
                   strstr(line,"<tag k='openGeoDB:name' v='")!=NULL){
                    // secondary name source
                    strcpy(linename,line);
                    name = extract_valuec(linename,"v");
                    normalname = 1;
                    foundname = 1;
                }
                if(normalname < 1 &&
                    strstr(line,"<tag k='name:")!=NULL){
                    // tertiary name source
                    strcpy(linename,line);
                    name = extract_valuec(linename,"v");
                    foundname = 1;
                }
                if(strstr(line,"tag k='population' v='") != NULL){
                    population = extract_valuei(line,"v");
                    foundpopulation = 1;
                }
    }
    //falls kein Name gefunden sollte Punkt nicht verzeichnet werden
    if(name == NULL)
        foundplace = foundpeak = 0;
    else
        htmlreplace(name);

    if(foundpopulation && foundplace && !foundcol)
        //d.h. ein Ort mit Einwohnerzahlangabe ist ausgelesen wurden
        if(population > 0){
            //Klassifizierung nach Einwohnerzahl ist sicherer...
            if(0 && population>100000)
                add_points(metropolis,k,name,float(population));
            else{
                if(population>10000)
                    add_points(city,k,name,float(population));
                else{
                    if(population>3000 || !strcmp(attr,"town"))
                    add_points(town,k,name,float(population));
                    else
                        if(population>50)
                            add_points(village,k,name,float(population));
                        else
                            add_points(locality,k,name,float(population));

                }
            }
        }
    if(foundplace && !foundpopulation  && !foundcol){
            //...als Klassifizierung nach OSM-tags, da keine Einwohnerzahl
            if(!strcmp(attr,"city"))
                add_points(metropolis,k,name,500000);
            else{
                if(!strcmp(attr,"town"))
                    add_points(town,k,name,5000);
                else{
                    if(   !strcmp(attr,"suburb")
                        ||!strcmp(attr,"village"))
                        add_points(village,k,name,500);
                    else
                        if(  !strcmp(attr,"hamlet")
                           ||!strcmp(attr,"locality")
                           ||!strcmp(attr,"alpine_hut")
                           ||!strcmp(attr,"isolated dwelling"))
                            add_points(locality,k,name,20);
                }
            }
    }
    if(foundcol && foundname){
        add_points(col,k,name,20);
    }
    if(foundpeak && foundname && !foundcol)
        add_points(peak,k,name,0);
}*/

void OSM::relationhandling(FILE *f,char* line,char* lineattr,char* linename,
                           qint64* currentseek){

	int relationlength = 0;
	int iswood = 0, iswater = 0;
    qint64 tID;
    Way* tw;
    while(strstr(line,"</relation>") == NULL){
        if(strstr(line,"<member type='way'") != NULL){
			relationlength++;
			tID = extract_valuei(line,"ref");
            tw = search(tID,*(D.waysavl));
            //if(tw != NULL && tw->type == 6)
            //	iswood = 1;
		}
        if(strstr(line,"<tag k='landuse' v='forest'/>") != NULL
        || strstr(line,"<tag k='natural' v='wood'/>") != NULL)
			iswood = 1;
        if(strstr(line,"<tag k='waterWay' v='riverbank'/>") != NULL
        || strstr(line,"<tag k='landuse' v='reservoir'/>") != NULL
        || strstr(line,"<tag k='natural' v='water'/>") != NULL)
			iswater = 1;
		fgets(line,999,f);
		quotationreplace(line);
	}
						
	if ((iswood || iswater) && relationlength > 0){
        relation* r =new relation;
		r->memberscount = relationlength;
        r->members = new Way*[relationlength];
        _fseeki64(f,*currentseek,SEEK_SET);
		fgets(line,999,f);
		quotationreplace(line);

		relationlength = -1;
        while(strstr(line,"</relation>") == NULL){
			fgets(line,999,f);
			quotationreplace(line);
            if(strstr(line,"<member type='way'") != NULL){
				relationlength++;
                r->members[relationlength] = new Way;
				r->members[relationlength]->id = extract_valuei(line,"ref");
			}
		}
		if(!iswater){
			add_relation(woodr,r);
			r->type = 6;
		}
		else{
			add_relation(waterr,r);
			r->type = 7;
		}
	}
}
void OSM::Wayhandling(FILE *f,char* line,char* lineattr,char* linename,
                      qint64* currentseek){
	char* attr = &line[0];
	char* name = NULL;
    int Waylength = 0;
    while(strstr(line,"</way>") == NULL){
        if(strstr(line,"<nd ref=") != NULL) Waylength++;
		fgets(line,999,f);
		quotationreplace(line);
        if(name == NULL && strstr(line,"<tag k='name' v='") != NULL){
            // take name only if no ref tag
			strcpy(linename,line);
			name = extract_valuec(linename,"v");
			htmlreplace(name);
		}
        if(strstr(line,"<tag k='ref' v='") != NULL){
            // take re
            strcpy(linename,line);
            name = extract_valuec(linename,"v");
            htmlreplace(name);
        }
	}
						
    if (Waylength > 0){
        _fseeki64(f,*currentseek,SEEK_SET);	//fur id neu lesen
		fgets(line,999,f);
		quotationreplace(line);
        Way* k = insert_AVL(extract_valuei(line,"id"),D.waysavl);
        k->nodec = Waylength;
        k->node = new KOO*[Waylength];
		k->type = 98;					//falls sich erweisen sollte, dass der
                                        //Way ungenutzt bleibt, muss er beim
										//speichern der fertig berechneten
										//Daten verworfen werden
        Waylength = -1;
        while(strstr(line,"</way>") == NULL){
			fgets(line,999,f);
			quotationreplace(line);
			if(strstr(line,"<nd ref=") != NULL){
                Waylength++;
                k->node[Waylength] = new KOO;
                k->node[Waylength]->id = extract_valuei(line,"ref");
			}
            if(strstr(line,"<tag k='highway' v='") != NULL ||
               strstr(line,"<tag k='construction' v='") != NULL){
                k->type = qMin(k->type,(char)5);
                strcpy(lineattr,line);
                attr = extract_valuec(lineattr,"v");
                if(!strcmp(attr,"motorway")||
                   !strcmp(attr,"motorway_link")||
                   !strcmp(attr,"trunk")||
                   !strcmp(attr,"trunk_link"))
                    k->type = qMin(k->type,(char)0);
                if(!strcmp(attr,"primary")||
                   !strcmp(attr,"primary_link"))
                    k->type = qMin(k->type,(char)1);
                if((!strcmp(attr,"secondary")
                    ||!strcmp(attr,"secondary_link")))
                    k->type = qMin(k->type,(char)2);
                if(!strcmp(attr,"tertiary")
                    ||!strcmp(attr,"tertiary_link"))
                    k->type = qMin(k->type,(char)3);
                if(!strcmp(attr,"residential")
                    ||!strcmp(attr,"unclassified")
                    ||!strcmp(attr,"road")
                    ||!strcmp(attr,"construction")
                    ||!strcmp(attr,"service"))
                    k->type = qMin(k->type,(char)4);
                if( !strcmp(attr,"cycleway")||
                    // !strcmp(attr,"path")||
                    // !strcmp(attr,"footway")||
                    0)//!strcmp(attr,"track"))
                    k->type = qMin(k->type,(char)5);
            }
            if(strstr(line,"<tag k='surface' v='") != NULL){
                strcpy(lineattr,line);
                attr = extract_valuec(lineattr,"v");
                if(!strcmp(attr,"asphalt")||
                   !strcmp(attr,"concrete")||
                   !strcmp(attr,"concrete:lanes")||
                   !strcmp(attr,"concrete:plates")||
                   !strcmp(attr,"paving_stones ")||
                   !strcmp(attr,"paved"))
                    k->type = qMin(k->type,(char)4);
                if(!strcmp(attr,"compacted")||
                   !strcmp(attr,"fine_gravel ")||
                   !strcmp(attr,"sett"))
                    k->type = (char)5;
                if(!strcmp(attr,"cobblestone:flattened"))
                    k->type = qMin(k->type,(char)4);
                if(!strcmp(attr,"cobblestone"))
                    k->type = qMin(k->type,(char)4);
            }

            if(strstr(line,"<tag k='tracktype' v='") != NULL){
                strcpy(lineattr,line);
                attr = extract_valuec(lineattr,"v");
                if(!strcmp(attr,"grade1"))
                    k->type = qMin(k->type,(char)4);
                if(!strcmp(attr,"grade2"))
                    k->type = (char)5;
                if(!strcmp(attr,"grade3")||
                   !strcmp(attr,"grade4")||
                   !strcmp(attr,"grade5"))
                   k->type = (char)5;
                   //k->type = 98;

            }
            if(strstr(line,"<tag k='tunnel' v='yes'/>") != NULL
            || strstr(line,"<tag k='bridge' v='") != NULL){
                add_line(NoHeight,k,name);
            }
            if((strstr(line,"<tag k='landuse' v='forest'/>") != NULL
             || strstr(line,"<tag k='natural' v='wood'/>") != NULL))
                k->type = qMin(k->type,(char)6);
            if((strstr(line,"<tag k='waterway' v='riverbank'/>") != NULL
            ||  strstr(line,"<tag k='landuse' v='reservoir'/>") != NULL
            ||  strstr(line,"<tag k='natural' v='water'/>") != NULL))
                k->type = qMin(k->type,(char)7);
            if(strstr(line,"<tag k='waterway' v='river'/>") != NULL)
                k->type = qMin(k->type,(char)8);
            if(strstr(line,"<tag k='waterway' v='stream'/>") != NULL)
                k->type = qMin(k->type,(char)9);
		}
        switch(k->type){
            case 0: add_line(motorWay,k,name); break;
            case 1: add_line(highway,k,name); break;
            case 2: add_line(primary,k,name); break;
            case 3: add_line(secondary,k,name); break;
            case 4: add_line(track,k,name); break;
            case 5: add_line(track2,k,name); break;
            case 6: add_line(wood,k,name); break;
            case 7: add_line(water,k,name); break;
            case 8: add_line(river,k,name); break;
            case 9: add_line(creek,k,name); break;
            default:;
        }
	}
}
/*
void OSM::Wayhandling(FILE *f,char* line,char* lineattr,char* linename,
                      qint64* currentseek){
    //SKIKEMAP wayhandling
    char* attr = &line[0];
    char* name = NULL;
    short namestate = -1;				//-1: nicht vorhanden, 0: zu lang fuer
                                        //Strasse, 1: nehmen,2: wieder loeschen
    int Waylength = 0;
    int foundWay = 0;
    while(strstr(line,"</way>") == NULL){
        if(strstr(line,"<nd ref=") != NULL) Waylength++;
        fgets(line,999,f);
        quotationreplace(line);
        if(  strstr(line,"<tag k='name' v='") != NULL
           ||strstr(line,"<tag k='ref' v='") != NULL){
            strcpy(linename,line);
            name = extract_valuec(linename,"v");
            htmlreplace(name);
            if(strlen(name)>0)	namestate = 1;
            //else				namestate = 0;
        }
    }

    if (Waylength > 0){
        _fseeki64(f,*currentseek,SEEK_SET);	//fur id neu lesen
        fgets(line,999,f);
        quotationreplace(line);
        Way* k = insert_AVL(extract_valuei(line,"id"),D.waysavl);
        k->nodec = Waylength;
        k->node = new KOO*[Waylength];
        k->type = 98;					//falls sich erweisen sollte, dass der
                                        //Way ungenutzt bleibt, muss er beim
                                        //speichern der fertig berechneten
                                        //Daten verworfen werden
        Waylength = -1;
        while(strstr(line,"</way>") == NULL){
            fgets(line,999,f);
            quotationreplace(line);
            if(strstr(line,"<nd ref=") != NULL){
                Waylength++;
                k->node[Waylength] = new KOO;
                k->node[Waylength]->id = extract_valuei(line,"ref");
            }
            if(strstr(line,"<tag k='highway' v='") != NULL){
                strcpy(lineattr,line);
                attr = extract_valuec(lineattr,"v");
                if(!strcmp(attr,"primary")||
                   !strcmp(attr,"primary_link")||
                   !strcmp(attr,"motorway")||
                   !strcmp(attr,"motorway_link")||
                   !strcmp(attr,"trunk")||
                   !strcmp(attr,"trunk_link"))
                    k->type = qMin(k->type,(char)0);
                if((!strcmp(attr,"secondary")
                    ||!strcmp(attr,"secondary_link")))
                    k->type = qMin(k->type,(char)2);
                if(!strcmp(attr,"tertiary"))
                   k->type = qMin(k->type,(char)3);
                if(!strcmp(attr,"residential")
                    ||!strcmp(attr,"road")
                    ||!strcmp(attr,"service"))
                    k->type = qMin(k->type,(char)4);
                if( !strcmp(attr,"unclassified")||
                    !strcmp(attr,"cycleway")||
                    !strcmp(attr,"track")||
                    !strcmp(attr,"footway"))
                    k->type = qMin(k->type,(char)5);
            }
            if(strstr(line,"<tag k='tracktype' v='") != NULL
                &&!foundWay){
                strcpy(lineattr,line);
                attr = extract_valuec(lineattr,"v");

                if(!strcmp(attr,"grade1"))
                    k->type = qMin(k->type,(char)4);
                if(!strcmp(attr,"grade2"))
                    k->type = qMin(k->type,(char)5);
                if(!strcmp(attr,"grade3"))
                    k->type = qMin(k->type,(char)1);
                if(!strcmp(attr,"grade4"))
                    k->type = 98;
                if(!strcmp(attr,"grade5"))
                    k->type = 98;
            }
            if(strstr(line,"<tag k='tunnel' v='yes'/>") != NULL
            || strstr(line,"<tag k='bridge' v='") != NULL){
                add_line(NoHeight,k,name);
            }
            if((strstr(line,"<tag k='landuse' v='forest'/>") != NULL
             || strstr(line,"<tag k='natural' v='wood'/>") != NULL)
            &&!foundWay){
                add_line(wood,k,name);
                k->type = 6;
                foundWay++;
            }
            if((strstr(line,"<tag k='waterway' v='riverbank'/>") != NULL
            ||  strstr(line,"<tag k='landuse' v='reservoir'/>") != NULL
            ||  strstr(line,"<tag k='natural' v='water'/>") != NULL)
            &&!foundWay){
                add_line(water,k,name);
                k->type = 7;
                foundWay++;
            }
            if(strstr(line,"<tag k='waterway' v='river'/>") != NULL
            &&!foundWay){
                add_line(river,k,name);
                k->type = 8;
                foundWay++;
            }

            if(strstr(line,"<tag k='waterway' v='stream'/>") != NULL
            &&!foundWay){
                add_line(creek,k,NULL);
                namestate = 2;
                k->type = 9;
                foundWay++;
            }
            if(strstr(line,"<tag k='surface' v='") != NULL){
                strcpy(lineattr,line);
                attr = extract_valuec(lineattr,"v");
                if(!strcmp(attr,"cobblestone")||
                   !strcmp(attr,"cobblestone:flattened"))
                    k->type = qMin(k->type,(char)1);
                if(!strcmp(attr,"asphalt")||
                   !strcmp(attr,"concrete")||
                   !strcmp(attr,"concrete:lanes")||
                   !strcmp(attr,"concrete:plates")||
                   !strcmp(attr,"paving_stones ")||
                   !strcmp(attr,"paved"))
                    k->type = qMin(k->type,(char)4);
                if(!strcmp(attr,"compacted")||
                   !strcmp(attr,"fine_gravel ")||
                   !strcmp(attr,"sett"))
                    k->type = qMin(k->type,(char)5);
            }
            switch(k->type){
                case 0: add_line(motorWay,k,name); break;
                case 5: add_line(highway,k,name); break;
                case 2: add_line(primary,k,name); break;
                case 3: add_line(secondary,k,name); break;
                case 4: add_line(track,k,name); break;
                case 1: add_line(track2,k,name); break;
            }
        }
    }
}
*/
void OSM::add_line(lines** l,Way* Way,char* name){
    lines* a = new lines;
    a->way = Way;
	if(*l == NULL) a->next = NULL;
	else a->next = *l;
	*l = a;
	if(name != NULL){
        Way->name = a->name = new char[strlen(name)+1];
		strcpy(a->name,name);
	}
	else
        a->name = Way->name = NULL;
}
points* OSM::add_points(points** l,KOO* KOO,char* name, float index){
    points* a =new points;
	a->index = index;
    a->k = KOO;
    a->k->status = 3;
    a->name = new char[strlen(name)+1];
	strcpy(a->name,name);
	if(*l == NULL){
		/*leere Liste, erstes Element*/
		a->next = NULL;
		a->prev = NULL;
		*l = a;
	}
	else{
		if((*l)->index < index){
			/*an Listenanfang einfuegen*/
			a->next = *l;
			a->prev = NULL;
			(*l)->prev = a;
			*l = a;
		}
		else{
			/*Liste durchgehen bis zur richtigen Position*/
			points* p = *l;
			while(p->next != NULL && p->index >= index)
				p = p->next;
			if(p->index >= index){
				/*Anfuegen an das Listenende*/
				p->next = a;
				a->prev = p;
				a->next = NULL;
			}
			else{
				/*Einfuegen mitten in die Liste*/
				p->prev->next = a;
				a->prev = p->prev;
				p->prev = a;
				a->next = p;
			}			
		}
	}
	return a;
}
void OSM::add_relation(relation** l, relation* r){
	r->next = *l;
	*l = r;
}
void OSM::invert_points(points** l){
	points* b = *l;
	points* p = *l;
	while(b != NULL){
		p = b;
		b = p->next;
		p->next = p->prev;
		p->prev = b;
	}
	*l = p;
}
void OSM::RelationstoFace(relation** l,lines** f){
	relation* r = *l;
	short *used, exists,foundneighbourmember;
    used = new short;
	lines* cf;
	int i,j,k,cm,cn;
    Way** Ways = new Way*;
    Way* grandWay;
    Way* tempWay;
	int nodecount;
	while(r != NULL){
        delete used;
		KOO *e,*s;
        used = new short[r->memberscount];
		for(i = 0; i < r->memberscount; i++){
			used[i] = 0;			
            tempWay = search(r->members[i]->id,*(D.waysavl));
            if(tempWay != NULL && tempWay->type != 99){
				/*der Weg existiert und kennt alle seine nodes*/
                r->members[i] = tempWay;
				used[i] = 0;
			}
			else
				used[i] = 1;
		}
		for(i = 0; i < r->memberscount; i++)
			if(!used[i]){
				if(r->members[i]->node[0					  ]->id == 
                   r->members[i]->node[r->members[i]->nodec-1]->id){
					/*falls der Member geschlossen ist, dann pruefen, ob er in
					f schon existiert, falls nicht und der Weg noch keinen Typ
					hat soll er in f eingefuegt werden*/
					exists = 0;
					cf = *f;
					while(cf != NULL){
                        if(cf->way->id == r->members[i]->id){
							exists = 1;
							cf = NULL;
						}
						else
							cf = cf->next;
					}
					if(!exists && r->members[i]->type > 8){
						add_line(f,r->members[i],NULL);
					}
					used[i] = 1;
				}
				else{
					/*falls der Member nicht geschlossen ist muss versucht
					werden, mit Hilfe der anderen Member eine geschlossene
					kurve zu konzentrieren. Members, die dabei helfen, liegen
                    im Array Ways*/
                    delete(Ways);
                    Ways = new Way*[r->memberscount];
					cm = 0;
					e = r->members[i]->node[0];
                    s = r->members[i]->node[r->members[i]->nodec-1];
                    Ways[cm]=r->members[i];
					used[i] = 1;
					do{
						foundneighbourmember = 0;
						for(j = 0; j < r->memberscount; j++){
							if(!used[j]
							&&(  s->id == r->members[j]->node[0]->id
                               ||s->id == r->members[j]->node[r->members[j]->nodec-1]->id)
							   ){
								/*falls der Weg j an den cm-Weg anknuepft*/
								cm++;
                                Ways[cm] = r->members[j];
								if(s->id == r->members[j]->node[0]->id)
                                    s = r->members[j]->node[r->members[j]->nodec-1];
								else
									s = r->members[j]->node[0];
								used[j] = 1;
								j = r->memberscount;
								foundneighbourmember = 1;
							}
						}
					}while(    foundneighbourmember
                            && (e->id != Ways[cm]->node[0]				  ->id)
                            && (e->id != Ways[cm]->node[Ways[cm]->nodec-1]->id));
					/*falls es gelungen ist, Members zu einer geschlossenen
					Kurve zu gruppieren, soll nun diese Kurve erstellt werden
					und in f abgelegt werden*/
					if(foundneighbourmember){
                        /*suche nach Weg unter Ways, der exklusiv von Relation
						genutzt wird*/
						int FoundExclusive = 1;
						int CheckMax = 6; if(l == woodr) CheckMax = 8;
						for(j = 0; j <= cm; j++){
							FoundExclusive = 1;
							for(int k = 0; k < CheckMax; k++){
								cf = *(ExclCheck[k]);
								while(cf != NULL){
                                    if(Ways[j]->id == cf->way->id){
										FoundExclusive = -1;
										cf = NULL;
										k = CheckMax;
									}
									else
										cf = cf->next;
								}
							}
                            if(FoundExclusive == 1){
								FoundExclusive = j;
								j = cm + 1;
							}
						}
						if(FoundExclusive > -1){
							/*es gibt einen exklusiven Weg der zum verknuepfen
							genutzt werden kann*/
							nodecount = 0;
							for(j = 0; j <= cm; j++)
                                nodecount = nodecount + Ways[j]->nodec;
							nodecount = nodecount - cm;
                            grandWay = new Way;

                            /*grandWay intialisieren*/
                            grandWay->balance = r->members[i]->balance;
                            grandWay->id = Ways[FoundExclusive]->id;
                            grandWay->nodec = nodecount;
                            grandWay->node =new KOO*[nodecount];
                            grandWay->type = r->type;

							cn = 0;
							s = r->members[i]->node[0];

                            grandWay->maxlat = -90;
                            grandWay->minlat = 90;
                            grandWay->maxlon = -180;
                            grandWay->minlon = 180;
							for(j = 0; j <= cm; j++){
                                if(Ways[j]->node[0]->id == s->id){
									/*vorwaerts fuellen*/
                                    for(k = 0; k < Ways[j]->nodec-1; k++){
                                        grandWay->node[cn] = Ways[j]->node[k];
										cn++;
									}
                                    s = Ways[j]->node[Ways[j]->nodec-1];
								}
								else{
									/*rueckwaerts fuellen*/
                                    for(k = Ways[j]->nodec-1; k > 0; k--){
                                        grandWay->node[cn] = Ways[j]->node[k];
										cn++;
									}
                                    s = Ways[j]->node[0];
								}
								/*umrandendes Rechteck bestimmen*/
                                if(Ways[j]->maxlat > grandWay->maxlat )
                                    grandWay->maxlat = Ways[j]->maxlat;
                                if(Ways[j]->minlat < grandWay->minlat )
                                    grandWay->minlat = Ways[j]->minlat;
                                if(Ways[j]->maxlon > grandWay->maxlon )
                                    grandWay->maxlon = Ways[j]->maxlon;
                                if(Ways[j]->minlon < grandWay->minlon )
                                    grandWay->minlon = Ways[j]->minlon;
							}
                            grandWay->node[grandWay->nodec-1]=grandWay->node[0];
                            add_line(f,grandWay,NULL);
						}
					}
				}
			}
		r = r->next;
	}
    delete(used);
}
void OSM::reset_KOO_status(KOO* k,char i){
	if(k != NULL){
		k->status = i;
		reset_KOO_status(k->l,i);
		reset_KOO_status(k->r,i);
	}
}
void OSM::calc_brect(){
    maxlat = -90;
    minlat = 90;
    maxlon = -180;
    minlon = 180;
    calc_brect(motorWay	);
    calc_brect(highway	);
	calc_brect(primary	); 
	calc_brect(secondary); 
	calc_brect(track	); 
	calc_brect(track2	); 
	calc_brect(river	); 
	calc_brect(creek	); 
	calc_brect(wood		); 
	calc_brect(water	);
	calc_brect(metropolis);	
	calc_brect(city);			
	calc_brect(town);			
	calc_brect(village);
    calc_brect(locality);
    calc_brect(col);
	calc_brect(peak);
}
void OSM::calc_brect(points** l){
	points* p = *l;
	while(p != NULL){
        if(p->k->lat > maxlat) maxlat = p->k->lat;
        if(p->k->lat < minlat) minlat = p->k->lat;
        if(p->k->lon > maxlon) maxlon = p->k->lon;
        if(p->k->lon < minlon) minlon = p->k->lon;
		p = p->next;
	}
}
void OSM::calc_brect(lines** l){
	lines* p = *l;
	while(p != NULL){
        for(int i = 0; i < p->way->nodec; i++){
            if(p->way->node[i]->lat > maxlat) maxlat = p->way->node[i]->lat;
            if(p->way->node[i]->lat < minlat) minlat = p->way->node[i]->lat;
            if(p->way->node[i]->lon > maxlon) maxlon = p->way->node[i]->lon;
            if(p->way->node[i]->lon < minlon) minlon = p->way->node[i]->lon;
		}
		p = p->next;
	}
}
Way* OSM::get_first_Way(){
    return *(D.waysavl);
}
KOO* OSM::get_first_KOO(){
	return *(D.nodes);
}
Neighbour* OSM::add_Neighbour(KOO* k, KOO** node, KOO* nb, Way* w, float d,
						int KOOc, readdirection rd){
    Neighbour *n;
	if( k->cd->neighbours == NULL){
        k->cd->neighbours =new Neighbour;
		n = k->cd->neighbours;
		n->next = NULL;
	}
	else{
		n = k->cd->neighbours;
        while(n->next != NULL && (n->nb != nb || n->way != w))
			/*Eine Nachbarbeziehung ist dann identisch, wenn die 2 Kreuzungen
			und der benutzte Weg übereinstimmen*/
			n = n->next;

        if(n->nb != nb || n->way != w){
			/*Die Nachbarbeziehung war ueberhaupt noch nicht vorhanden*/
            n->next =new Neighbour;
			n = n->next;
			n->next = NULL;
		}
		else{
			if(d < n->d){
				/*Die Nachbarbeziehung war mit einer laengeren Distanz
				vorhanden und sollte ueberschrieben werden*/
				if(n->tp != NULL){
                    delete(n->tp);
					n->tp = NULL;
					n->brthr->tp = NULL;
				}
				
			}
			else
				/*die bestehende Nachbarbezieung hat einen kuerzeren Weg,
				einfuegen abbrechen*/
				return NULL;
		}
	}
	n->d = d;
	n->hm = 0;
    n->t = d;
	n->nb = nb;
	n->brthr = NULL;
    n->way = w;
	n->nodec = KOOc;
	n->node = node;
	n->rd = rd;
	n->tpc = 0;
	n->tp = NULL;
	k->status = 2;
	return n;
}
double OSM::getminlat(){
	return minlat;
}
double OSM::getminlon(){
	return minlon;
}
double OSM::getmaxlat(){
	return maxlat;
}
double OSM::getmaxlon(){
	return maxlon;
}

void OSM::calc_peak_index_pro(HeightData* h){
	points* p = *peak;
	int wc = 50, j, i, ac = 10;
	float h0;
	float d = float(0.2);
	
    float* index = new float[ac];
    float* index2 = new float[ac];
	float is,is2;
	float Ahigher;
    qint64**c = new qint64*[ac];
	for(i = 0; i < ac; i++)
        c[i] = new qint64[wc];
    float mperlon, mperlat;
	while(p != NULL){
        mperlon = 2*M_PI*6371/360*cos(p->k->lat/180.*M_PI) * 1000/1200;
        mperlat = 2*M_PI*6371/360 * 1000/1200;
        h0 = h->height_t(p->k->lon,p->k->lat,NULL,0);
        h->paek_pro(p->k->lon,p->k->lat,c,&Ahigher,wc,ac,d,mperlon,mperlat);
		for(i = 0; i < ac; i++){
			index[i] = 0;
			index2[i] = 0;
			/*Berechnet Volumen*/
			if( h0 > c[i][0])
				for(j = 1; j < wc; j++)
					if(c[i][j]<c[i][j-1])
						index[i] = index[i] + (h0 - c[i][j]);
					else
						j = wc;
			/*Berechnet Bergformel*/
			for(j = 0; j < wc; j++)
				if( h0 > c[i][j]
					&&(h0-c[i][j])*(h0-c[i][j])/(i*d*1000) > index2[i])
					index2[i] = float((h0-c[i][j])*(h0-c[i][j])/((0.5+j)*d*1000));
		}
		is = 0;
		is2 = 0;
		p->index = 0;
		for(j = 0; j < ac; j++){
			is  = is  + sqrt(index[j]);
			is2 = is2 + index2[j];
		}
		p->index = sqrt(is*is*is2/ac);
		p->index = float(p->index * max(1.,1+(0.002-Ahigher)*500));

		p = p->next;
	}
	
	for(i = 0; i < ac; i++)
        delete(c[i]);
    delete(c);
    delete(index);
    delete(index2);
}
void OSM::sort_peaks(){
	short change = 1;
	points p,*l = *peak;
	if(l != NULL){
		p.next = l;
		l->prev = &p;
		FILE* f;
		f = fopen("pointssortiert.txt","w");
		while(change){
			change = 0;
			l = p.next;
			while(l->next != NULL){
				/*a-b-c-d ==> a-c-b-d*/
				if(l->index < l->next->index){
					l->prev->next = l->next;//a->c
					l->next->prev = l->prev;//a<-c
					l->prev	= l->next;		//c<-b
					if(l->next->next != NULL){
						l->next->next->prev = l;//b<-d
						l->next = l->next->next;//b->d;
					}
					else
						l->next = 0;
					l->prev->next = l;		//c->b;
					change = 1;
				}
				else
					l = l->next;
			}
		}
		l = p.next;
		while(l != NULL){
            fprintf(f,"%f - %s - %f N %f E\n",l->index,l->name,l->k->lat,l->k->lon);
			l = l->next;
		}

		fclose(f);

		p.next->prev = NULL;				//Abkoppeln vom Pufferelement
		*peak = p.next;						//Listenanfang setzen
	}
}
void OSM::save(char* filename, double minlat, double maxlat, 
			   double minlon, double maxlon){
	FILE* f;
	f = fopen(filename,"wb");
	if(	dmaxlat == -90
		&& dminlat == 90
		&& dmaxlon == -180
		&& dminlon == 180){
		dmaxlat = maxlat;
		dminlat = minlat;
		dmaxlon = maxlon;
		dminlon = minlon;
	}

	fwrite(&minlat,sizeof(double),1,f);
	fwrite(&maxlat,sizeof(double),1,f);
	fwrite(&minlon,sizeof(double),1,f);
	fwrite(&maxlon,sizeof(double),1,f);

    save_lines(motorWay	,f,minlat,maxlat,minlon,maxlon);
    save_lines(highway	,f,minlat,maxlat,minlon,maxlon);
	save_lines(primary	,f,minlat,maxlat,minlon,maxlon); 
	save_lines(secondary,f,minlat,maxlat,minlon,maxlon); 
	save_lines(track	,f,minlat,maxlat,minlon,maxlon); 
	/*lines* b = NULL;
	save_lines(&b		,f,minlat,maxlat,minlon,maxlon);*/
	save_lines(track2	,f,minlat,maxlat,minlon,maxlon); 
	save_lines(river	,f,minlat,maxlat,minlon,maxlon); 
	save_lines(creek	,f,minlat,maxlat,minlon,maxlon); 
	save_lines(wood		,f,minlat,maxlat,minlon,maxlon); 
	save_lines(water	,f,minlat,maxlat,minlon,maxlon);

	save_points(metropolis,	f,minlat,maxlat,minlon,maxlon);	
	save_points(city,		f,minlat,maxlat,minlon,maxlon);			
	save_points(town,		f,minlat,maxlat,minlon,maxlon);			
	save_points(village,	f,minlat,maxlat,minlon,maxlon);
    save_points(locality,	f,minlat,maxlat,minlon,maxlon);
    save_points(col,    	f,minlat,maxlat,minlon,maxlon);
	
	/*Bergliste rückwärts speichern, da sie sortiert durch insert_points
	geladen werden soll*/
	invert_points(peak);
	save_points(peak,		f,minlat,maxlat,minlon,maxlon);
	invert_points(peak);

	/*speichere Kreuzungsdaten ab, vorher Kreuzungen durchzählen, nachher
	diese Anzahl bei currentseek speichern*/
    int c = 0;
    qint64 currentseek = _ftelli64(f);
    fwrite(&c,sizeof(int),1,f);
	save_cross(*D.nodes, f, &c, minlat, maxlat, minlon, maxlon);
    _fseeki64(f,currentseek,SEEK_SET);
    fwrite(&c,sizeof(int),1,f);
	
	fclose(f);
	reset_KOO_status(*D.nodes);
}
void OSM::save_lines(lines** l, FILE* f, double minlat, double maxlat, 
					double minlon, double maxlon){
    int c = 0;
	int len;
	if(*l == NULL)
        fwrite(&c,sizeof(int),1,f);
	else{
		lines*  k = *l;
		while(k != NULL){
            if(isin(k->way,minlat,maxlat,minlon,maxlon))
				c++;
			k = k->next;
		}
        fwrite(&c,sizeof(int),1,f);
		k = *l;
		while(k != NULL){
            if(isin(k->way,minlat,maxlat,minlon,maxlon)){
                save_Way(k->way,f);
				len = 0;
				if(k->name == NULL){
					fwrite(&len,sizeof(int),1,f);
				}
				else{
					len = strlen(k->name);
					fwrite(&len,sizeof(int),1,f);
					fwrite(k->name,sizeof(char),len,f);
				}
			}
			k = k->next;
		}
	}
}
void OSM::save_points(points** p, FILE* f, double minlat, double maxlat, 
					double minlon, double maxlon){
    int c = 0;
	int len;
	if(*p == NULL)
        fwrite(&c,sizeof(int),1,f);
	else{
		points*  k = *p;
		while(k != NULL){
            if(isin(k->k,minlat,maxlat,minlon,maxlon))
				c++;
			k = k->next;
		}
        fwrite(&c,sizeof(int),1,f);
		k = *p;
		while(k != NULL){
            if(isin(k->k,minlat,maxlat,minlon,maxlon)){
                save_node(k->k,f);
				fwrite(&(k->index),sizeof(float),1,f);
				len = 0;
				if(k->name == NULL){
					fwrite(&len,sizeof(int),1,f);
				}
				else{
					len = strlen(k->name);
					fwrite(&len,sizeof(int),1,f);
					fwrite(k->name,sizeof(char),len,f);
				}
			}
			k = k->next;
		}
	}
}
void OSM::save_node(KOO* k, FILE* f){
    fwrite(&(k->id),sizeof(qint64),1,f);
	fwrite(&(k->lon),sizeof(double),1,f);
	fwrite(&(k->lat),sizeof(double),1,f);
	if(k->cd != NULL)
		/*markiere den Node als noch zu speichernde Kreuzung*/
		k->status = 10;						
}
void OSM::save_Way(Way* w, FILE* f){
    fwrite(&(w->id),sizeof(qint64),1,f);
	fwrite(&(w->type),sizeof(char),1,f);
    fwrite(&(w->nodec),sizeof(int),1,f);
    for(int i = 0; i < w->nodec; i++)
		save_node(w->node[i],f);
}
void OSM::save_cross(KOO* k, FILE* f, int* c, double minlat,
					 double maxlat, double minlon, double maxlon){
	if(k != NULL){
		if(k->status == 10
		   && isin(k,dminlat,dmaxlat,dminlon,dmaxlon)){
			(*c)++;
            fwrite(&(k->id),sizeof(qint64),1,f);
			/*Anzahl i der Nachbarkreuzungen bestimmen*/
			short i = 0;
            Neighbour* n = k->cd->neighbours;
			while(n != NULL){
				/*gespeichert werden immer nur die VorwärtsNachbarrelationen,
				sonst ist das gespeichert redundant. Beim Laden werden dann
				beider Kreuzungen befuellt werden*/
				if(n->nb->status == 10 && n->rd == fw
				   &&isin(n->nb,dminlat,dmaxlat,dminlon,dmaxlon))
					i++;				
				n = n->next;
			}
			fwrite(&i,sizeof(short),1,f);

			n = k->cd->neighbours;
			while(n != NULL){
				if(n->nb->status == 10 && n->rd == fw
				   && isin(n->nb,dminlat,dmaxlat,dminlon,dmaxlon)){
					fwrite(&(n->rd),		sizeof(readdirection),1,f);
					fwrite(&(n->d),			sizeof(float),1,f);
                    fwrite(&(n->hm),		sizeof(float),1,f);
                    fwrite(&(n->nb->id),	sizeof(qint64),1,f);
                    fwrite(&(n->way->id),	sizeof(qint64),1,f);
					fwrite(&(n->nodec),		sizeof(int),1,f);
					int j = 0;
                    for(int l = n->way->nodec-1; l >=0; l--){
                        if(n->way->node[l] == k)
							j = l;			//findet zugehoerigen Index in der 
					}						//KOO-Liste des Weges
					fwrite(&j,sizeof(int),1,f);
					fwrite(&(n->tpc),		sizeof(int),1,f);
					fwrite(n->tp,			sizeof(trackpoint),n->tpc,f);
				}
				n = n->next;
			}
		}
		/*Rekursionsaufruf*/
		save_cross(k->l,f,c, minlat, maxlat, minlon, maxlon);
		save_cross(k->r,f,c, minlat, maxlat, minlon, maxlon);
	}
}
void OSM::reset_KOO_status(KOO* k){
	if(k != NULL){
		if(k->status == 10)
			k->status = 2;
		/*rekursiver Aufruf*/
		reset_KOO_status(k->l);
		reset_KOO_status(k->r);
	}
}
void OSM::load(char* filename){
	FILE* f;
	f = fopen(filename,"rb");
    if(f != NULL){
        double minlat, maxlat, minlon, maxlon;
        fread(&minlat,sizeof(double),1,f);
        fread(&maxlat,sizeof(double),1,f);
        fread(&minlon,sizeof(double),1,f);
        fread(&maxlon,sizeof(double),1,f);

        if(minlat < this->minlat) this->minlat = minlat;
        if(maxlat > this->maxlat) this->maxlat = maxlat;
        if(minlon < this->minlon) this->minlon = minlon;
        if(maxlon > this->maxlon) this->maxlon = maxlon;
        load_lines(motorWay	,f);
        load_lines(highway	,f);
        load_lines(primary	,f);
        load_lines(secondary,f);
        load_lines(track	,f);
        load_lines(track2	,f);
        load_lines(river	,f);
        load_lines(creek	,f);
        load_lines(wood		,f);
        load_lines(water	,f);
        load_points(metropolis,	f);
        load_points(city,		f);
        load_points(town,		f);
        load_points(village,	f);
        load_points(locality,	f);
        load_points(col,    	f);
        load_points(peak,		f);
        int c;
        fread(&c,sizeof(int),1,f);
        for(int i = 0; i < c; i++)
            load_cross(f);
        fclose(f);
        //QMessageBox(QMessageBox::NoIcon,"OSMende","equal heights follow...").exec();

        equal_crossheight(get_first_KOO());
        //QMessageBox(QMessageBox::NoIcon,"OSMende","equal heights finished.").exec();
        status = 3;
    }
}
KOO* OSM::load_node(FILE* f){
    qint64 id;
    fread(&id,sizeof(qint64),1,f);
	KOO* k = insert_AVL(id,D.nodes);
	if(k->lat == 200){
		fread(&(k->lon),sizeof(double),1,f);
		fread(&(k->lat),sizeof(double),1,f);
		k->status = 0;
		return k;
	}
	else{
		fseek(f,2*sizeof(double),SEEK_CUR);
		return k;
	}	
}
Way* OSM::load_Way(FILE* f){
    qint64 id;
    int length,i;
	char type;
    Way* w;
    fread(&id,sizeof(qint64),1,f);
	fread(&type,sizeof(char),1,f);
    fread(&length,sizeof(int),1,f);
    w = insert_AVL(id,D.waysavl);
    if(w->type == -1 || w->nodec < length){
		/*lesen und fuellen, da noch nicht vorhanden oder kürzer*/
        KOO** newKOO = new KOO*[length];
		for(i = 0; i < length; i++)
			newKOO[i] = load_node(f);

		if(w->type != -1){
            replace_Way_KOO(w,newKOO,length);
			w = NULL;					//Der Weg war schon vorhanden, wurde
										//nur geaendert, deshalb return NULL
										//da OSM::load_lines damit arbeitet
		}
		else{
			w->type = type;
            w->nodec = length;
			w->node = newKOO;
			w->maxlat = -90;
			w->minlat =  90;
			w->maxlon =-180; 
			w->minlon = 180;
			for(i = 0; i < length; i++){
				if(w->node[i]->lat > w->maxlat) w->maxlat = float(w->node[i]->lat);
				if(w->node[i]->lat < w->minlat) w->minlat = float(w->node[i]->lat);
				if(w->node[i]->lon > w->maxlon) w->maxlon = float(w->node[i]->lon);
				if(w->node[i]->lon < w->minlon) w->minlon = float(w->node[i]->lon);
			}
		}
	}
	else{
        fseek(f,length*(2*sizeof(double)+sizeof(qint64)),SEEK_CUR);
		w = NULL;						//auf keinen Fall neues Line
	}
	return w;
}
void OSM::load_cross(FILE* f){
	readdirection rd;
	float d,hm;
    qint64 id, nid, wid;
	int j,tpc,nodec;
	short c;
    Neighbour *n1, *n2;
	trackpoint *tp;
    fread(&id, sizeof(qint64),1,f);
	fread(&c,sizeof(short),1,f);
    Way* w;
	KOO* k = search(id,*D.nodes), *kn;
	if(k != NULL){
		if(k->cd == NULL){
			/*Kreuzungsdaten anlegen, falls noch nicht vorhanden*/
            k->cd = new crossdata;
			k->cd->neighbours = NULL;
            k->cd->s = new STRONG[2];
			k->cd->s->from = NULL;
			k->cd->s->via = NULL;
		}
		
		/*Alle Kreuzungsabgänge bearbeiten*/
		for(short i = 0; i < c; i++){
			/*einlesen*/

			fread(&(rd),	sizeof(readdirection),1,f);
			fread(&(d),		sizeof(float),1,f);
			fread(&(hm),	sizeof(float),1,f);
            fread(&(nid),	sizeof(qint64),1,f);
            fread(&(wid),	sizeof(qint64),1,f);
			fread(&(nodec),	sizeof(int),1,f);
			fread(&j,		sizeof(int),1,f);
			fread(&(tpc),	sizeof(int),1,f);
			
			kn = search(nid,*D.nodes);
            w = search(wid,*D.waysavl);
			
            if(kn != NULL && w != NULL && j+(nodec-1) < w->nodec){
				/*Achtung: bei verschieden alte OSM-DAtensätze gleichen sich
				die Wege nicht unbedingt aufgrund der Gleichheit ihrer ID.
				Im Falle der verschiedenheit muss wenigstens gewährleistet,
                dass Neighbour nicht über das nodearray hinausliest*/
				if(kn->cd == NULL){
					/*Kreuzungsdaten anlegen, falls noch nicht vorhanden*/
                    kn->cd = new crossdata;
					kn->cd->neighbours = NULL;
                    kn->cd->s = new STRONG[2];
					kn->cd->s->from = NULL;
					kn->cd->s->via = NULL;
				}
                n1 = add_Neighbour(k ,&(w->node[j]),kn,w,d,nodec,fw);
                n2 = add_Neighbour(kn,&(w->node[j]),k ,w,d,nodec,bw);
				if(n1 != NULL && n2 != NULL){
					/*Normalfalll: hin und rückweg existieren noch nicht*/
                    tp = new trackpoint[tpc];
					fread(tp,sizeof(trackpoint),tpc,f);
					n1->tp = tp;
					n1->tpc = tpc;
					n1->brthr = n2;
					n1->hm = hm;
					n2->tp = tp;
					n2->tpc = tpc;
					n2->brthr = n1;
                    n2->hm = qMax(float(0),hm - (tp[tpc-1].h - tp[0].h));
					//n2->hm = (n1->hm+n2->hm)/2.;
					//n1->hm = n2->hm;
				}
				else
					if(n1 != NULL){
						/*Ringstrasse bspw.*/
                        tp = new trackpoint[tpc];
						fread(tp,sizeof(trackpoint),tpc,f);
						n1->tp = tp;
						n1->tpc = tpc;
						n1->brthr = n1;
						n1->hm = hm;
					}
					else{
						/*Straßen existieren bereits, seek setzen statt
						weiter auslesen*/
                        del_Neighbour(k,n1);
                        del_Neighbour(kn,n2);
						fseek(f,tpc*sizeof(trackpoint),SEEK_CUR);
					}
			}
			else{
				fseek(f,tpc*sizeof(trackpoint),SEEK_CUR);
			}
		}
	}
	else
		/*sonst kann die Kreuzung nicht angelegt werden, sollte aber nicht 
		passieren*/
		for(short i = 0; i < c; i++){
			fseek(f,  (  sizeof(readdirection)
						+2*sizeof(float)
                        +2*sizeof(qint64)
						+2*sizeof(int)),SEEK_CUR);
			fread(&(tpc),	sizeof(int),1,f);
			fseek(f,tpc*sizeof(trackpoint),SEEK_CUR);
		}
}
void OSM::load_lines(lines** l, FILE* f){
    int c;
	int len;
    Way* w;
	char name[1000];
    fread(&c,sizeof(int),1,f);
    for(int i = 0; i < c; i++){
        w = load_Way(f);
		fread(&len,sizeof(int),1,f);
		if(len > 0 && w != NULL){
            /*Way noch nicht vorhanden und besitzt Namen*/
			fread(name,sizeof(char),len,f);
			name[len]='\0';
			add_line(l,w,name);
		}
		else
			if(w != NULL)
                /*Way noch nicht vorhanden, aber kein Name*/
				add_line(l,w,NULL);
			else
				/*weg bereits vorhanden*/
				fseek(f,len*sizeof(char),SEEK_CUR);
	}
}
void OSM::load_points(points** p, FILE* f){
    int c;
	int len;
	float index;
	KOO* k;
	char name[1000];
    fread(&c,sizeof(int),1,f);
    for(int i = 0; i < c; i++){
		k = load_node(f);
		fread(&index,sizeof(float),1,f);
		fread(&len,sizeof(int),1,f);
		fread(name,sizeof(char),len,f);
		if(len > 0 && k->status == 0){
			name[len] = '\0';
			add_points(p,k,name,index);
		}
		else
			if(k->status == 0)
				add_points(p,k,NULL,index);
	}
}
void OSM::check_lines(){
    check_lines_rek(motorWay);
    check_lines_rek(highway	);
	check_lines_rek(primary	); 
	check_lines_rek(secondary); 
	check_lines_rek(track	); 
	check_lines_rek(track2	); 
	check_lines_rek(river	); 
	check_lines_rek(creek	); 
	check_lines_rek(wood	); 
	check_lines_rek(water	);
}
void OSM::check_lines_rek(lines** l){
	lines dummy, *t = &dummy, *b;
	dummy.next = *l;
	while(t->next != NULL){
        if(t->next->way->type >= 98){
			/*Element loeschen*/
			b = t->next;
			t->next = t->next->next;
            delete(b);
		}
		else
			t = t->next;
	}
	*l = dummy.next;
}
void OSM::quotationreplace(char* a){
	int i = 0;
	while(a[i] != '\0'){
		if(a[i] == '\"')
			a[i] = '\'';
		i++;
	}
}
void OSM::htmlreplace(char* c){
	if(c != NULL){
		char html[2][6] = {{'&','#','3','9',';','\0'},{'&','#','3','4',';','\0'}};
		char newchar[2] = {39,34};
		for(int i = 0; i < 2; i++){
			while(strstr(c,html[i]) != NULL){
				char* ct = strstr(c,html[i]);
				ct[0] = newchar[i];
				strcpy(&ct[1],&ct[strlen(html[i])]);
			}
		}
	}
}
void OSM::select_cross(double lat, double lon, char mintype, char maxtype,
                       KOO** kdest, float* d,KOO* ksrc){
    if(ksrc != NULL){
        if(ksrc->status == 2){
            bool reachable= false;
            Neighbour* n = ksrc->cd->neighbours;
            while(n != NULL && ! reachable){
                if(n->way->type <= maxtype && n->way->type >= mintype)
                    reachable = true;
                n = n->next;
            }
            if(reachable){
                float d2 = distance(lat,lon,ksrc->lat,ksrc->lon);
                if(d2 < *d){
                    *d = d2;
                    *kdest = ksrc;
                }
            }
        }
        select_cross(lat,lon,mintype,maxtype,kdest,d,ksrc->l);
        select_cross(lat,lon,mintype,maxtype,kdest,d,ksrc->r);
    }
}

void OSM::select_cross(double lat, double lon, KOO** kdest, float* d,KOO* ksrc){
	if(ksrc != NULL){
		if(ksrc->status == 2){
			float d2 = distance(lat,lon,ksrc->lat,ksrc->lon);
			if(d2 < *d){
				*d = d2;
				*kdest = ksrc;
			}
		}
		select_cross(lat,lon,kdest,d,ksrc->l);
		select_cross(lat,lon,kdest,d,ksrc->r);
	}
}
Way* OSM::select_Way(double lat, double lon){
    KOO *k1 = get_first_KOO(), *k2 = NULL;
    Way k;
	k.minlat = float(lat - 0.0001);
	k.maxlat = float(lat + 0.0001);
	k.minlon = float(lon - 0.00015);
	k.maxlon = float(lon + 0.00015);
	lines dummy, *l;
	dummy.next = NULL;
    extract_Way(&dummy,get_first_Way(),&k,0,8);
	
    Way* wmin = NULL;
	float dmin = 10000, d1, d2;
	l = dummy.next;
	while(l != NULL){
		d1 = 1000;
		d2 = 1000;
        for(int i = 0; i < l->way->nodec; i++)
            if(distance(l->way->node[i]->lat,l->way->node[i]->lon,lat,lon) < d1){
				d2 = d1;
                d1 = distance(l->way->node[i]->lat,l->way->node[i]->lon,lat,lon);
				k2 = k1;
                k1 = l->way->node[i];
			}
		if((d1 + d2) - distance(k1->lat, k1->lon, k2->lat, k2->lon) < dmin){
			dmin = (d1 + d2) - distance(k1->lat, k1->lon, k2->lat, k2->lon);
            wmin = l->way;
		}
		l = l->next;
	}
	
	

	/*Liste löschen*/
	while(dummy.next != NULL){
		l = dummy.next;
		dummy.next = l->next;
        delete(l);
	}

	return wmin;
}
void OSM::extract_Way(lines* l, Way* w, Way* k, char mintype, char maxtype){
	if(w != NULL){
		if(w->type <= maxtype && w->type >= mintype
			&& isin(w,k->minlat,k->maxlat,k->minlon,k->maxlon)){
            lines* tl = new lines;
            tl->way = w;
			tl->next = l->next;
			l->next = tl;
		}
        extract_Way(l,w->l,k,mintype,maxtype);
        extract_Way(l,w->r,k,mintype,maxtype);
	}
}
void OSM::change_Waytype(Way* w, char newtype){
	lines** l = NULL;
	switch(w->type){
        case 0: l = motorWay; break;
        case 1: l = highway; break;
		case 2: l = primary; break;		
		case 3: l = secondary; break;	
		case 4: l = track; break;		
		case 5: l = track2; break;
		case 6: l = wood; break;		
		case 7: l = water; break;	
		case 8: l = river; break;
		case 9: l = creek; break;	
	}
	if(l != NULL){
		char* name = NULL;
		del_lines(w,l,name);
		switch(newtype){
            case 0: l = motorWay; break;
            case 1: l = highway; break;
			case 2: l = primary; break;		
			case 3: l = secondary; break;	
			case 4: l = track; break;		
			case 5: l = track2; break;
			case 6: l = wood; break;		
			case 7: l = water; break;	
			case 8: l = river; break;
			case 9: l = creek; break;	
		}
		add_line(l,w,name);
	}
	w->type = newtype;
}
void OSM::del_lines(Way* w, lines** l, char* keepname){
	if(*l != NULL){
		lines* tl = *l, *tlbuffer;
        while(tl ->next != NULL && tl->next->way != w)
			tl = tl->next;				//Weg raussuchen
		if(tl->next != NULL){
			/*falls gefunden, dann aus der Liste entfernen*/
			tlbuffer = tl->next;
			tl->next = tl->next->next;
			keepname = tlbuffer->name;
            delete(tlbuffer);
		}
		else{
			/*falls es das erste Listenelement ist*/
            if((*l)->way == w){
				tlbuffer = (*l);
				*l = tlbuffer->next;
				keepname = tlbuffer->name;
                delete(tlbuffer);
			}
		}
	}
}
void OSM::del_Way(Way* w){
	lines** l = NULL;
	switch(w->type){
        case 0: l = motorWay; break;
        case 1: l = highway; break;
		case 2: l = primary; break;		
		case 3: l = secondary; break;	
		case 4: l = track; break;		
		case 5: l = track2; break;
		case 6: l = wood; break;		
		case 7: l = water; break;	
		case 8: l = river; break;
		case 9: l = creek; break;
		default: l = creek; break;
	}
	char *name = NULL;
	del_lines(w,l,name);
	if(name != NULL)
        delete(name);

    for(int i = 0; i < w->nodec; i++){
		if(w->node[i]->cd != NULL){
            del_Neighbour(w->node[i],w);
		}
	}
	w->type = 99;
}
void OSM::del_Neighbour(KOO* k, Way* w){
	if(k->cd != NULL){
        Neighbour dummy, *n = &dummy, *nbuffer;
		dummy.next = k->cd->neighbours;
		while(n->next != NULL){
            if(n->next->way == w){
				/*Die Nachbarrelation n->next muss geloescht werden*/
				nbuffer = n->next;
				n->next = nbuffer->next;
                delete(nbuffer);
			}
			else
				n = n->next;
		}
		k->cd->neighbours = dummy.next;
	}
}
void OSM::del_Neighbour(KOO* k, Neighbour* n){
	if(n != NULL && k != NULL && k->cd != NULL){
        Neighbour dummy, *tn = &dummy;
		char found = 0;
		dummy.next = k->cd->neighbours;
		while(tn->next != NULL){
			if(tn->next == n){
				tn->next = tn->next->next;
				found = 1;
			}
			else
				tn = tn->next;
		}
		if(found)
            delete(n);
		k->cd->neighbours = dummy.next;
	}
}
short OSM::check_AVL(Way* w){
	if(w != NULL){
		if(w->l != NULL && w->l->id > w->id)
			return 0;
		if(w->r != NULL && w->r->id < w->id)
			return 0;
		return check_AVL(w->l)*check_AVL(w->r);
	}
	else
		return 1;
}
void OSM::replace_Way_KOO(Way* w, KOO** newKOO, int newlength){
	KOO** KOObuffer = w->node;
    int lengthbuffer = w->nodec;
	w->node = newKOO;
    w->nodec = newlength;
	for(int i = 0; i < lengthbuffer; i++)
		/*alle Wegpunkte untersuchen, ob sie Kreuzungen sind*/
		if(KOObuffer[i]->cd != NULL){
            Neighbour* nb = KOObuffer[i]->cd->neighbours;
			short deletenow = 0;
			while(nb != NULL){
				/*alle Nachbarsschaftsbeziehung der Kreuzung untersuchen*/
                if(nb->way == w){
					/*die Nachbarschaftsbeziehung verläuft auf dem zu
					veraendernden Weg w*/
					int KOOcount = 0;
					short countnow = 0;
					for(int j = 0; j < newlength; j++){
						/*Pruefe, ob die alten KOO auf dem neuem Weg vorhanden
						sind*/
						if(countnow)
							KOOcount++;
                        if(w->node[j] == KOObuffer[i]){
							if(!countnow){
								countnow = 1;
								nb->node = &(w->node[j]);
								nb->rd = fw;
								nb->brthr->node = &(w->node[j]);
								nb->brthr->rd = bw;
							}
							else{
								countnow = 0;
								nb->nodec = KOOcount + 1;
								nb->brthr->nodec = KOOcount + 1;
							}
                        }
						if(w->node[j] == nb->nb){
							if(!countnow){
								countnow = 1;
								nb->node = &(w->node[j]);
								nb->rd = bw;
								nb->brthr->node = &(w->node[j]);
								nb->brthr->rd = fw;
							}
							else{
								countnow = 0;
								nb->nodec = KOOcount + 1;
								nb->brthr->nodec = KOOcount + 1;
							}
						}
					}
					if(countnow || !KOOcount)
						/*wenigstens einer der Kreuzungen ist im neuem Weg
						nicht vorhanden. Diese Nachbarschaftsbeziehung muss
						aufgegeben werden*/
						deletenow = 1;
						
				}
				if(deletenow){
                    Neighbour* nbbuffer = nb;
					nb = nb->next;
                    del_Neighbour(nbbuffer->nb,nbbuffer->brthr);
                    del_Neighbour(KOObuffer[i],nbbuffer);
					deletenow = 0;
				}
				else
					nb = nb->next;
			}
		}
    delete(KOObuffer);
}
void OSM::equal_crossheight(KOO* k){
	if(k != NULL){
        if(k->cd != NULL && k->cd->neighbours != NULL){
            Neighbour* n = k->cd->neighbours;
			short equal = 1;
			double h1,h2;
			if(n->rd == fw) h1 = n->tp[0       ].h;
			else			h1 = n->tp[n->tpc-1].h;
			/*Gibt es überhaupt Diskrepanzen bezueglich der Kreuzungshoehe*/
            while(n->next != NULL && equal == 1){
                if(n->next->rd == fw) h2 = n->next->tp[0             ].h;
                else				  h2 = n->next->tp[n->next->tpc-1].h;
				if(h1 != h2) equal = 0;
                h1 = h2;
				n = n->next;
			}
			/*Falls ja, sollten die Diskrepanzen beseitigt werden*/
            if(!equal){
				double avh = 0;			//Average height
                int nbc = 0;			//Neighbour count
				n = k->cd->neighbours;
				/*Alle abgehenden Wege sollen die Duchschnittshoehe der
				Kreuzung an der Kreuzung aufzeigen*/
				while(n != NULL){
                    if(n->rd == fw) avh = avh + n->tp[0             ].h;
                    else			avh = avh + n->tp[n->tpc-1].h;
					nbc++;
					n = n->next;
				}
				avh = 1.*avh/nbc;
				/*Korriere alle abgehenden Wege der Kreuzung linear*/
				double dh;
				n = k->cd->neighbours;
				while(n != NULL){
					if(n->d > 0){
						if(n->rd == fw){
							dh = n->tp[0].h - avh;
                            for(int i = 0; i < n->tpc; i++){
								n->tp[i].h = float(n->tp[i].h 
											 - dh* (n->d - n->tp[i].d)/n->d);
							}
						}
						else{
							dh = n->tp[n->tpc-1].h - avh;
                            for(int i = 0; i < n->tpc; i++){
								n->tp[i].h = float(n->tp[i].h - dh* (n->tp[i].d)/n->d);
							}
						}
					}
					else{
						if(n->rd == fw) n->tp[0].h = float(avh);
						else	 n->tp[n->tpc-1].h = float(avh);
					}
					n = n->next;
				}
			}
		}
		equal_crossheight(k->l);
		equal_crossheight(k->r);
	}
}
void OSM::normalize_Hm(KOO* k){
    if(k != NULL){
        if(k->cd != NULL && k->cd->neighbours != NULL){
            Neighbour* n = k->cd->neighbours;
            while(n != NULL){
                if(n->rd == fw){
                    n->hm = (n->hm + n->brthr->hm)/2.;
                    n->brthr->hm = n->hm;
                }
                n->d = n->tp[n->tpc-1].d;
                n = n->next;
            }
        }
        normalize_Hm(k->l);
        normalize_Hm(k->r);
	}
}
void OSM::invert_Hm(KOO* k){
    if(k != NULL){
        if(k->cd != NULL && k->cd->neighbours != NULL){
            Neighbour* n = k->cd->neighbours;
            float hmswap;
            while(n != NULL){
                if(n->rd == fw){
                    hmswap = n->hm;
                    n->hm = n->brthr->hm;
                    n->brthr->hm = hmswap;
                }
                n = n->next;
            }
        }
        invert_Hm(k->l);
        invert_Hm(k->r);
	}
}
void OSM::swap_Hm_d(KOO* k){
    if(k != NULL){
        if(k->cd != NULL && k->cd->neighbours != NULL){
            Neighbour* n = k->cd->neighbours;
            float hmswap;
            while(n != NULL){
                hmswap = n->hm;
                n->hm = n->d*10+0.00001;
                n->d = hmswap;
                n = n->next;
            }
        }
        swap_Hm_d(k->l);
        swap_Hm_d(k->r);
    }
}
void OSM::reset_Hm(KOO* k){
    if(k != NULL){
        if(k->cd != NULL && k->cd->neighbours != NULL){
            Neighbour* n = k->cd->neighbours;
            while(n != NULL){
                n->brthr->hm = n->hm = 0;
                if(n->rd == fw){
                    for(int i = 1; i < n->tpc; i++)
                        if(n->tp[i].d > n->tp[i-1].d){
                            if(n->tp[i].h > n->tp[i-1].h)
                                n->hm += (n->tp[i].h - n->tp[i-1].h);
                            else
                                n->brthr->hm -= (n->tp[i].h - n->tp[i-1].h);
                        }
                }
                n = n->next;
            }
        }
        reset_Hm(k->l);
        reset_Hm(k->r);
    }
}

void OSM::points_Hm(KOO* k){
    if(k != NULL){
        if(k->cd != NULL && k->cd->neighbours != NULL){
            Neighbour* n = k->cd->neighbours;
            while(n != NULL){
                n->brthr->hm = n->hm = 0;
                if(n->rd == fw){
                    for(int i = 1; i < n->tpc; i++)
                        if(n->tp[i].d > n->tp[i-1].d){
                            if(n->tp[i].h > n->tp[i-1].h)
                                n->hm += (n->tp[i].h - n->tp[i-1].h)
                                        *(n->tp[i].h - n->tp[i-1].h)
                                        /(n->tp[i].d - n->tp[i-1].d)/1000;
                            else
                                n->brthr->hm += (n->tp[i].h - n->tp[i-1].h)
                                               *(n->tp[i].h - n->tp[i-1].h)
                                               /(n->tp[i].d - n->tp[i-1].d)/1000;
                        }
                }
                n = n->next;
            }
        }
        points_Hm(k->l);
        points_Hm(k->r);
    }
}
void OSM::refresh_time(KOO* k, float (*t)(Neighbour* n, void* inst),void* buf){
    if(k != NULL){
        if(k->cd != NULL && k->cd->neighbours != NULL){
            Neighbour* n = k->cd->neighbours;
            while(n != NULL){
                n->t = t(n,buf);
                n = n->next;
            }
        }
        refresh_time(k->l,t,buf);
        refresh_time(k->r,t,buf);
    }
}

void OSM::merge(KOO* knew, KOO* kold){
	if(	knew != NULL && kold != NULL &&
		knew->cd != NULL && kold->cd != NULL &&
		knew->cd->neighbours != NULL && kold->cd->neighbours != NULL){
		double newh;
		if(knew->cd->neighbours->rd == fw)
			newh = knew->cd->neighbours->tp[0].h;
		else
			newh = knew->cd->neighbours->tp[knew->cd->neighbours->tpc-1].h;
        Neighbour* tn = kold->cd->neighbours;
		while(tn != NULL){
			KOO* klast;
			if(tn->rd == fw)	klast = tn->node[0];
			else				klast = tn->node[tn->nodec-1];
			double dd =  distance(klast->lat,klast->lon,knew->lat,knew->lon)
						-distance(klast->lat,klast->lon,kold->lat,kold->lon);
			tn->d = tn->brthr->d = tn->d + dd;
			if(tn->rd == fw){
				for(int i = 1; i < tn->tpc; i++)
					tn->tp[i].d = tn->tp[i].d+dd;
				tn->node[0] = knew;
				tn->tp[0].h = newh;
			}
			else{
				tn->tp[tn->tpc-1].d = tn->tp[tn->tpc-1].d + dd;
				tn->node[tn->nodec-1] = knew;
				tn->tp[tn->tpc-1].h = newh;
			}
			tn->brthr->nb = knew;
			double hmfw= 0, hmbw = 0;
			for(int i = 1; i < tn->tpc; i++)
				if(tn->tp[i].h > tn->tp[i-1].h) 
					hmfw = hmfw + tn->tp[i].h - tn->tp[i-1].h;
				else
					hmbw = hmbw - tn->tp[i].h + tn->tp[i-1].h;

			if(tn->rd == fw){
				tn->hm = hmfw;
				tn->brthr->hm = hmbw;
			}
			else{
				tn->hm = hmbw;
				tn->brthr->hm = hmfw;
			}

			if(tn->nb == knew && tn != tn->brthr){
                del_Neighbour(knew,tn->brthr);
				tn = tn->next;
			}
			else{
                Neighbour* tb = tn->next;
				tn->next = knew->cd->neighbours;
				knew->cd->neighbours = tn;
				tn = tb;
			}

		}
		kold->cd = NULL;
		kold->status = 0;
	}
}
void OSM::deleteRect(double minlat, double maxlat, 
					 double minlon, double maxlon){
	/*alle Wege löschen*/
    del_Way(get_first_Way(),minlat,maxlat,minlon,maxlon);
	/*alle Punkte löschen*/
    points **pt[7];
	pt[0] = metropolis;		
	pt[1] = city;			
	pt[2] = town;			
	pt[3] = village;
    pt[4] = locality;
    pt[5] = col;
    pt[6] = peak;
    for(int i = 0; i < 7; i++){
		points* p = *pt[i];
		points pbuf; pbuf.next = p;
		if(p != NULL) p->prev = &pbuf;
		while(p != NULL){
			if(isin(p,minlat,maxlat,minlon,maxlon)){
				p->prev->next = p->next;
				if(p->next != NULL)
					p->next->prev = p->prev;
			}
			p = p->next;
		}
		if(pbuf.next != NULL)
			pbuf.next->prev = NULL;
		*pt[i] = pbuf.next;	
	}
}
void OSM::del_Way(Way* w, double minlat, double maxlat,
					 double minlon, double maxlon){
	if(w != NULL){
		if(iscompletein(w,minlat,maxlat,minlon,maxlon))
            del_Way(w);
        del_Way(w->l,minlat,maxlat,minlon,maxlon);
        del_Way(w->r,minlat,maxlat,minlon,maxlon);
	}
}
void OSM::delete_AVL(Way* l){
    if(l != NULL){
        if(l->node != NULL) delete l->node;
        if(l->name != NULL) delete l->name;
        delete_AVL(l->l);
        delete_AVL(l->r);
        delete l;
    }
}

void OSM::delete_AVL(KOO* l){
    if(l != NULL){
        if(l->cd != NULL){
            Neighbour *n = l->cd->neighbours, *nt;
            while(n != NULL){
                nt = n;
                n = n->next;
                if(nt->rd == fw)
                    delete nt->tp;
                delete nt;
            }
            if(l->cd->s != NULL) delete l->cd->s;
            delete l->cd;
        }
        delete_AVL(l->l);
        delete_AVL(l->r);
        delete l;
    }
}

void OSM::delete_lines(lines* l){
    lines* lt;
    while(l != NULL){
        lt = l;
        l = l->next;
        delete lt;
    }
}

void OSM::delete_points(points* l){
    points* lt;
    while(l != NULL){
        lt = l;
        l = l->next;
        if(lt->name != NULL) delete lt->name;
        delete lt;
    }
}

void OSM::delete_relation(relation* l){
    relation* lt;
    while(l != NULL){
        lt = l;
        l = l->next;
        if(lt->members != NULL) delete l->members;
        delete lt;
    }
}
