#include "terrain.h"

Terrain::Terrain(OSM* o, Track** ct,int width, int height,config* c){
    settings = c;
    t = ct;
    heightLE.setText(QString("%1").arg(height));
    widthLE.setText(QString("%1").arg(width));
    superele.setText("1");
    fog.setText("20000");
    fogos.setText("15000");
    textureMpx.setText("2.0");

    pm = new QPixmap(width,height);
    info = new pxinfo*[width];
    //Img[0] = Img[1] = Img[2] = Img[3] = NULL;
    maps[0] = maps[1] = maps[2] = maps[3] = NULL;
    for(int i = 0; i < width; i++) info[i] = new pxinfo[height];

    h = new HeightData();
    h->set_SRTM3_folder(c->SRTM3folder);
    m = new MAP(h,o);
    m->settings(&(c->MS));

    QGridLayout* L = new QGridLayout();
    connect(&BtnRecalc,SIGNAL(clicked()),this,SLOT(recalc()));
    connect(&BtnRetexture,SIGNAL(clicked()),this,SLOT(retexture()));
    connect(&Btnsave,SIGNAL(clicked()),this,SLOT(save_png()));
    Pic.resize(width,height);
    //Pic->setSizePolicy(QSizePolicy::Expanding,QSizePolicy::Expanding);
    QScrollArea* sa = new QScrollArea;
    sa->setWidget(&Pic);
    L->addWidget(sa,0,0,1,6);
    LLatPos.setText("lat pos:");
    L->addWidget(&LLatPos,1,0);
    L->addWidget( & LatPos,1,1);
    LLonPos.setText("lon pos:");
    L->addWidget(&LLonPos,1,2);
    L->addWidget(  &LonPos,1,3);
    LhPos.setText("height pos:");
    L->addWidget(  &LhPos,1,4);
    L->addWidget(   & hPos,1,5);
    LelevLE.setText("elevation:");
    L->addWidget(&LelevLE,2,0);
    L->addWidget(  &elevLE,2,1);
    LazimutLE.setText("azimut:");
  L->addWidget(&LazimutLE,2,2);
  L->addWidget(  &azimutLE,2,3);
    LzoomLE.setText("zoom:");
    L->addWidget(&LzoomLE,2,4);
    L->addWidget(  &zoomLE,2,5);
    Lsuperele.setText("superelevation:");
    L->addWidget(&Lsuperele,3,0);
    L->addWidget(  &superele,3,1);
    Lfog.setText("fog distance:");
  L->addWidget(&Lfog,3,2);
  L->addWidget(  &fog,3,3);
    Lfogos.setText("fog offset:");
    L->addWidget(&Lfogos,3,4);
    L->addWidget(  &fogos,3,5);
    LwidthLE.setText("width:");
   L->addWidget(&LwidthLE,4,0);
   L->addWidget( &widthLE,4,1);
    LheightLE.setText("heigth:");
  L->addWidget(&LheightLE,4,2);
  L->addWidget(  &heightLE,4,3);
    L->addWidget(&Btnsave,4,4);
    L->addWidget(&BtnRecalc,4,5);

    Leminlat.setText("Texture minlat:");
    L->addWidget(&Leminlat,5,0);
    L->addWidget(  &eminlat,5,1);
    Lemaxlat.setText("Texture maxlat:");
    L->addWidget(&Lemaxlat,5,2);
    L->addWidget(  &emaxlat,5,3);
    Leminlon.setText("Texture minlon:");
    L->addWidget(&Leminlon,6,0);
    L->addWidget(  &eminlon,6,1);
    Lemaxlon.setText("Texture maxlon:");
    L->addWidget(&Lemaxlon,6,2);
    L->addWidget(  &emaxlon,6,3);
    LtextureMpx.setText("Texture Megapixel:");
      L->addWidget(&LtextureMpx,5,4);
      L->addWidget(  &textureMpx,5,5);
      L->addWidget(&BtnRetexture,6,5);

    setLayout(L);
    resize(width,height);
    BtnRecalc.setText("Render");
    BtnRetexture.setText("Retexture");
    Btnsave.setText("save png");
}
Terrain::~Terrain(){
    delete pm;
    for(int i = 0; i < pm->width(); i++) delete info[i];
    delete info;
    if(h != NULL) delete h;
    if(m != NULL) delete m;
    deletemaps();
}
void Terrain::retexture(){
    deletemaps();
    maxlat = emaxlat.text().toDouble();
    minlat = eminlat.text().toDouble();
    maxlon = emaxlon.text().toDouble();
    minlon = eminlon.text().toDouble();

    //prepare texture
    double dlat = (maxlat-minlat)*0.51;
    double dlon = (maxlon-minlon)*0.51;
    if(m != NULL){
        int Mappxcount = int(textureMpx.text().toDouble()*1e6);
        int imgpxs = int(sqrt(Mappxcount));
        m->resize(imgpxs,imgpxs);

        for(int i = 0; i < 2; i++)
            for(int j = 0; j < 2; j++){
                maps[i*2+j] = new MAP(NULL,NULL);
                maps[i*2+j]->resize(imgpxs,imgpxs);
                maps[i*2+j]->set_borders(minlat+i*dlat,minlat+(i+1)*dlat,
                                         minlon+j*dlon,minlon+(j+1)*dlon);
                maps[i*2+j]->calc_borders();

                m->set_borders(minlat+i*dlat,minlat+(i+1)*dlat,
                               minlon+j*dlon,minlon+(j+1)*dlon);
                m->calc_borders();
                m->draw_map(NULL,NULL,*t,0);

                Img[i*2+j] = m->get_Map()->toImage();
            }
    }
    recalc();
}
void Terrain::save_png(){
    QString filename = QFileDialog().getSaveFileName(this,"save 3D picture","","potable network graphics (*.png)");
    if(!filename.isEmpty() && pm != NULL)
        pm->save(filename,"PNG");
}

void Terrain::deletemaps(){
    for(int i = 0; i < 4; i++){
        //if(Img[i] != NULL) delete Img[i];
        if(maps[i] != NULL) delete maps[i];
    }
}

void Terrain::setpos(double lat, double lon, double h,
                     double azimut, double elev, double zoom){
    LatPos.setText(QString("%1").arg(lat));
    LonPos.setText(QString("%1").arg(lon));
    hPos.setText(QString("%1").arg(h));
    azimutLE.setText(QString("%1").arg(azimut));
    elevLE.setText(QString("%1").arg(elev));
    zoomLE.setText(QString("%1").arg(zoom));


    R0 = xyz(lat,lon,h);
    xyz nlon;
    nlon.fromEukl(-sin(lon*0.0174532925),cos(lon*0.0174532925),0);
    xyz nlat;
    nlat.fromEukl(-cos((90-lat)*0.0174532925)*cos(lon*0.0174532925),
                  -cos((90-lat)*0.0174532925)*sin(lon*0.0174532925),
                   sin((90-lat)*0.0174532925));
    xyz nr;
    nr.fromEukl(   sin((90-lat)*0.0174532925)*cos(lon*0.0174532925),
                   sin((90-lat)*0.0174532925)*sin(lon*0.0174532925),
                   cos((90-lat)*0.0174532925));

    nv = nlat*( sin(elev*0.0174532925)*cos(azimut*0.0174532925))
        +nlon*( sin(elev*0.0174532925)*sin(azimut*0.0174532925))
        +nr*   -cos(elev*0.0174532925);
    nh = nlat*( sin(azimut*0.0174532925))
        +nlon*(-cos(azimut*0.0174532925));
    nz = nlat*(-cos(elev*0.0174532925)*cos(azimut*0.0174532925))
        +nlon*(-cos(elev*0.0174532925)*sin(azimut*0.0174532925))
        +nr*   -sin(elev*0.0174532925);

    nv.normalize();
    nh.normalize();
    nz.normalize();

    double check[9];
    check[0]=nr*nlat;
    check[1]=nr*nlon;
    check[2]=nlon*nlat;
    check[3]=nv*nh;
    check[4]=nv*nz;
    check[5]=nh*nz;
    check[6]=nz*nlat;
    check[7]=nv*nr;
    check[8]=nh*nlon;

    this->elev = elev;
    this->zoom = zoom;

    minlat = lat - 0.5;
    maxlat = lat + 0.5;
    minlon = lon - 0.5;
    maxlon = lon + 0.5;
}
void Terrain::setpos(double lat1, double lon1, double h1,
                     double lat2, double lon2, double h2,
                     double zoom){
    setpos(lat1,lon1,h1,0,0,zoom);
    xyz n = R0.difference(xyz(lat2,lon2,h2));
    n.normalize();
    double elevnew = asin(nv*n)*180./3.14159;
    double azimutnew = atan2(nh*n,nz*n)*180./3.14159;
    elevLE.setText(QString("%1").arg(elevnew));
    azimutLE.setText(QString("%1").arg(azimutnew));
    setpos(lat1,lon1,h1, azimutnew, elevnew, zoom);
}

void raytracehelper(threadcontainer* tc,Terrain* t){
    t->raytrace4(tc->Img,tc->m,tc->q,tc->xstart,tc->xend);
}

void Terrain::summit_collection(double lat, double lon, double h,
                                QString filename, QString label, double azimut0){
    zoom = 2*1980;
    set_terrain(lat-0.7,lat+0.7,lon-0.7/cos(lat/180*3.14159),
                lon+0.7/cos(lat/180*3.14159));
    double dlat = (lat-minlat)*0.999;
    double dlon = (lon-minlon)*0.999;
    QImage Img[4];
    MAP* maps[4];
    if(m != NULL){
        int Mappxcount = 60e6;
        int imgpxs = int(sqrt(Mappxcount));
        m->resize(imgpxs,imgpxs);

        for(int i = 0; i < 2; i++)
            for(int j = 0; j < 2; j++){
                maps[i*2+j] = new MAP(NULL,NULL);
                maps[i*2+j]->resize(imgpxs,imgpxs);
                maps[i*2+j]->set_borders(minlat+i*dlat,lat+i*dlat,
                                         minlon+j*dlon,lon+j*dlon);
                maps[i*2+j]->calc_borders();

                m->set_borders(minlat+i*dlat,lat+i*dlat,
                               minlon+j*dlon,lon+j*dlon);
                m->calc_borders();
                m->draw_map(NULL,NULL,*t,0);

                Img[i*2+j] = m->get_Map()->toImage();
            }
    }
    //this->h->reset_t(minlat-dlat,maxlat+dlat,minlon-dlon,maxlon+dlon);
    int picnu  = -1;
    double azio = 0;
    double dazimut = 1;
    double az2 = -188;
    threadcontainer tc[16];
    QPainter q(pm);
    QFont qf("Helvetica ",15);
    q.setFont(qf);
    for(int i = 0; i < 16; i++){
        tc[i].h = new HeightData();
        tc[i].h->set_SRTM3_folder(settings->SRTM3folder);
        tc[i].h->reset_t(2*minlat-maxlat,
                         2*maxlat-minlat,
                         2*minlon-maxlon,
                         2*maxlon-minlon);
        tc[i].m = maps;
        tc[i].Img = Img;
        tc[i].xstart = 1./16.*i*pm->width();
        tc[i].xend = 1./16.*(i+1)*pm->width();
        tc[i].q = &q;
    }
    for(double azimut = azimut0; azimut <= azimut0; azimut+=dazimut){
        setpos(lat,lon,h,azimut,0,zoom);
        /*double radius = 2300*(pow(cos((azimut+210)/180*3.14159),2)*6+5+cos((azimut+30)/180*3.14159)*3);
        radius = (23000+4000*cos(azimut/180*3.14159)+6000*sin((azimut+90)/90*3.14159));
        radius = (23000+4000*cos(azimut/180*3.14159)+6000*sin((azimut+90)/90*3.14159)+5000*exp((azimut-10)*(10-azimut)/300));
        if(azimut < az2){
            radius *= exp((azimut-az2)*(az2-azimut)/2400);
            azio -=dazimut*0.5*qMin((az2-azimut)*0.02,1.);
            h += dazimut*3.7;
        }*/
        double radius = 22000;
        xyz pos = R0+nz*radius;
        setpos(pos.lat(),pos.lon(),pos.h(),azimut+180.,0,zoom);
        q.fillRect(0,0,pm->width(),pm->height(),Qt::black);
        QFutureSynchronizer<void> sync;
        //raytrace4(Img,maps,&q,0,pm->width());
        for(int i = 0; i < 16; i++){
            QFuture<void> future = QtConcurrent::run(raytracehelper,&(tc[i]),this);
            sync.addFuture(future);
        }
        sync.waitForFinished();
        picnu++;
        if(!label.isEmpty()){
            q.setPen(Qt::white);
            q.drawText(20,35,label);
        }
        pm->save(QString(filename).append(QString("%1.png")
                                          .arg(azimut,4,'f',0,QChar('0'))),"PNG");
    }
    for(int i = 0; i < 4; i++)
        delete maps[i];
    for(int i = 0; i < 8; i++)
        delete tc[i].h;
}

void Terrain::drawterrain(double minlat,double maxlat,
                          double minlon, double maxlon,
                          double resolution){
    newhdata(minlat,maxlat,minlon,maxlon,resolution);
    h->set_rect(minlat,maxlat,minlon,maxlon);
    h->interpolate_bicubic_conv3_mercator(lonsteps,latsteps,0,hc,dhdx,dhdy);
    add_DEM();
    QList<px> pxlist;

    float xo = pm->width() / 2;
    float yo = pm->height() / 20;
    int x,y;
    QPainter q(pm);
    for(int loni = 0; loni < lonsteps-0; loni++){
        double lon = minlon + (maxlon-minlon)*loni/(lonsteps-1);
        for(int lati = 0; lati < lonsteps-0; lati++){
            double lat = maxlat - (maxlat-minlat)*lati/(latsteps-1);
            px pxt(R0.difference(xyz(lat,lon,hc[loni][lati])));
            pxt.c = htocolor(lati,loni);
            pxt.hlon = loni;
            pxt.hlat = lati;
            pxt.normalize();
            pxt.project(&nh,&nv,&nz);
            if(pxt.Z > 0){
                pxt.zoom(zoom);
                pxlist.append(pxt);
            }
        }
    }
    qSort(pxlist.begin(),pxlist.end());
    //for(int i = 0; i < pxlist.length(); i++)
    //    pxref[pxlist[i].hlon][pxlist[i].hlon] = &(pxlist[i]);
    QPointF P[4];
    for(int i = 0; i < pxlist.length(); i++){
        q.setPen(QPen(pxlist[i].c,10,Qt::SolidLine,Qt::RoundCap,Qt::RoundJoin));
        //int lati = pxlist[i].hlat;
        //int loni = pxlist[i].hlon;

        //P[0].setX(pxref);
        //for(int dx = -1; dx < 3; dx+=2)
        //P[0].setX(pxlist[i].X+xo);
        //P[0].setY(pxlist[i].Y+yo);
        q.drawPoint(pxlist[i].X+xo,pxlist[i].Y+yo);
        //q.drawPolygon();
    }
    deletehdata();
}
void Terrain::terrain_raytrace(double minlat,double maxlat,
                          double minlon, double maxlon){
    set_terrain(minlat,maxlat,minlon,maxlon);
    QImage Img;
    if(m != NULL){
        int Mappxcount = textureMpx.text().toDouble()*1e6;
        double widthtoheight = (maxlon-minlon)*cos(M_PI*(maxlat+minlat)/2./180.)
                                /(maxlat-minlat);
        int height = int(sqrt(Mappxcount/widthtoheight));
        int width= int(height*widthtoheight);
        m->resize(width,height);
        m->set_borders(minlat,maxlat,minlon,maxlon);
        m->calc_borders();
        m->draw_map(NULL,NULL,*t,0);
        Img = m->get_Map()->toImage();
    }
    h->reset_t(minlat,maxlat,minlon,maxlon);
    raytrace(&Img);

    update();
}
void Terrain::raytrace(QImage *Img){
    double dloncorrection = 1./cos((minlat+maxlat)/360.*3.1415);
    double fmax = 300000;
    double df = 5;
    float xo = pm->width() / 2;
    float yo = pm->height() / 10;
    double f = 0;
    double lat,lon,ffw,fbw,Iloss;
    double lambdaI = fog.text().toDouble();
    double Io = fogos.text().toDouble();
    double superelev = superele.text().toDouble();
    xyz r;
    QPainter q(pm);
    q.fillRect(0,0,pm->width(),pm->height(),Qt::black);
    for(int x = 0; x < pm->width(); x++){
        f = 0;
        for(int y = pm->height()-1; y >= 0; y--){
            xyz n = nz + nh*((x-xo)/zoom) + nv*((y-yo)/zoom);
            n.normalize();
            // ray trace
            do{
                f = f+df;
                r = R0+(n*f);
            }while(r.h() > h->height_t_save(r.lon(),r.lat(),NULL,0)*superelev && f<fmax);
            // ray hit surface
            if(f < fmax){
                //refine ray length
                fbw = f-df;
                ffw = f+df;
                for(int i = 0; i < 5; i++){
                    f = (ffw+fbw)/2;
                    r = R0+(n*f);
                    if(r.h() > h->height_t_save(r.lon(),r.lat(),NULL,0)*superelev)
                        fbw = f;
                    else
                        ffw = f;
                }
                // paint px
                lon = r.lon();
                lat = r.lat();
                double ht =h->height_t_save(lon,lat,NULL,0);
                QColor c;
                if(m == NULL)
                    c = htocolor(ht,
                                 h->height_t_save(lon,lat,NULL,1)*dloncorrection,
                                 h->height_t_save(lon,lat,NULL,2));
                else
                    c = m->color(lon,lat,Img);

                Iloss=exp(-qMax(0.,f-Io)/lambdaI);
                c = QColor(c.red()*Iloss+140*(1-Iloss),
                           c.green()*Iloss+140*(1-Iloss),
                           c.blue()*Iloss+140*(1-Iloss));
                q.setPen(c);
                q.drawPoint(x,y);
                info[x][y].d = f;
                info[x][y].h = ht;
                info[x][y].lat = r.lat();
                info[x][y].lon = r.lon();
            }
        }

    }
}
void Terrain::raytrace4(QImage *Img, MAP **m){
    double fmax = 150000;
    double df = 5;
    float xo = pm->width() / 2;
    float yo = pm->height() / 10;
    double f = 0;
    double lat,lon,ffw,fbw,Iloss;
    double lambdaI = 12000;
    double Io = 12000;
    xyz r;
    QPainter q(pm);
    q.fillRect(0,0,pm->width(),pm->height(),Qt::black);
    for(int x = 0; x < pm->width(); x++){
        f = 0;
        for(int y = pm->height(); y >= 0; y--){
            xyz n = nz + nh*((x-xo)/zoom) + nv*((y-yo)/zoom);
            n.normalize();
            // ray trace
            do{
                f = f+df;
                r = R0+(n*f);
            }while(r.h() > h->height_t_save(r.lon(),r.lat(),NULL,0) && f<fmax);
            // ray hit surface
            if(f < fmax){
                //refine ray length
                fbw = f-df;
                ffw = f+df;
                for(int i = 0; i < 8; i++){
                    f = (ffw+fbw)/2;
                    r = R0+(n*f);
                    if(r.h() > h->height_t_save(r.lon(),r.lat(),NULL,0))
                        fbw = f;
                    else
                        ffw = f;
                }
                // paint px
                lon = r.lon();
                lat = r.lat();
                QColor c;
                int mapi = 0;
                do{ c = m[mapi]->color(lon,lat,&(Img[mapi]));
                    mapi++;
                }while(c == Qt::black && mapi<4);

                if(c != Qt::black){
                    Iloss=exp(-qMax(0.,f-Io)/lambdaI);
                    c = QColor(c.red()*Iloss+140*(1-Iloss),
                               c.green()*Iloss+140*(1-Iloss),
                               c.blue()*Iloss+140*(1-Iloss));
                    q.setPen(c);
                }
                else
                    q.setPen(QColor(140,140,140));
                q.drawPoint(x,y);

            }
        }
    }
}
void Terrain::raytrace4(QImage *Img, MAP **m,QPainter *q,int xstart, int xend){
    double fmax = 300000;
    double df = 5;
    float xo = pm->width() / 2;
    float yo = pm->height() / 10;
    double f = 0;
    double lat,lon,ffw,fbw,Iloss;
    double lambdaI = fog.text().toDouble();
    double Io = fogos.text().toDouble();
    double superelev = superele.text().toDouble();
    HeightData *h = new HeightData();
    h->set_SRTM3_folder(settings->SRTM3folder);
    h->reset_t(2*minlat-maxlat,
               2*maxlat-minlat,
               2*minlon-maxlon,
               2*maxlon-minlon);
    xyz r;
    for(int x = xstart; x < xend; x++){
        f = 0;
        for(int y = pm->height()-1; y >= 0; y--){
            xyz n = nz + nh*((x-xo)/zoom) + nv*((y-yo)/zoom);
            n.normalize();
            double red = 0, green = 0, blue = 0;
            int samples;
            // ray trace
            do{
                f = f+df;
                r = R0+(n*f);
            }while(r.h() > h->height_t_save(r.lon(),r.lat(),NULL,0)*superelev && f<fmax);
            // ray hit surface
            if(f < fmax){
                //refine ray length
                fbw = f-df;
                ffw = f+df;
                for(int i = 0; i < 8; i++){
                    f = (ffw+fbw)/2;
                    r = R0+(n*f);
                    if(r.h() > h->height_t_save(r.lon(),r.lat(),NULL,0))
                        fbw = f;
                    else
                        ffw = f;
                }
                // paint px
                lon = r.lon();
                lat = r.lat();
                QColor c;
                int mapi = 0;
                do{ c = m[mapi]->color(lon,lat,&(Img[mapi]));
                    mapi++;
                }while(c == Qt::black && mapi<4);

                if(c != Qt::black){
                    Iloss=exp(-qMax(0.,f-Io)/lambdaI);
                    c = QColor(c.red()*Iloss+140*(1-Iloss),
                               c.green()*Iloss+140*(1-Iloss),
                               c.blue()*Iloss+140*(1-Iloss));
                }
                else
                    c = QColor(140,140,140);

                qmutex.lock();
                q->setPen(c);
                q->drawPoint(x,y);
                qmutex.unlock();

                info[x][y].d = f;
                info[x][y].h = r.h();
                info[x][y].lat = r.lat();
                info[x][y].lon = r.lon();
            }
        }
    }
    delete h;
}

void Terrain::recalc(){
    for(int i = 0; i < pm->width();i++)
        delete info[i];
    delete info;
    delete pm;

    pm = new QPixmap(widthLE.text().toInt(),heightLE.text().toInt());
    info = new pxinfo*[pm->width()];
    for(int i = 0; i < pm->width(); i++) info[i] = new pxinfo[pm->height()];
    Pic.resize(pm->size());

    setpos(LatPos.text().toDouble(),
           LonPos.text().toDouble(),
           hPos.text().toDouble(),
           azimutLE.text().toDouble(),
           elevLE.text().toDouble(),
           zoomLE.text().toDouble());

    // prepare height data
    threadcontainer tc[16];
    QPainter q(pm);
    for(int i = 0; i < 16; i++){
        tc[i].h = new HeightData();
        tc[i].h->set_SRTM3_folder(settings->SRTM3folder);
        tc[i].h->reset_t(2*minlat-maxlat,
                         2*maxlat-minlat,
                         2*minlon-maxlon,
                         2*maxlon-minlon);
        tc[i].m = maps;
        tc[i].Img = Img;
        tc[i].xstart = 1./16.*i*pm->width();
        tc[i].xend = qMin(pm->width(),int(1./16.*(i+1)*pm->width()));
        tc[i].q = &q;
    }

    q.fillRect(0,0,pm->width(),pm->height(),Qt::black);
    QFutureSynchronizer<void> sync;
    for(int i = 0; i < 16; i++){
        QFuture<void> future = QtConcurrent::run(raytracehelper,&(tc[i]),this);
        sync.addFuture(future);
    }
    sync.waitForFinished();

    for(int i = 0; i < 8; i++)
        delete tc[i].h;
}

QPixmap* Terrain::getpixmap(){
    return pm;
}
QSize Terrain::sizeHint() const{
    return QSize(250,100);
}
QSize Terrain::minimumSizeHint() const{
    return QSize(100,100);
}
void Terrain::paintEvent(QPaintEvent *){
    Pic.setPixmap(*pm);
}
void Terrain::newhdata(double minlat,double maxlat,
                       double minlon, double maxlon,
                       double resolution){
    this->minlat=minlat;
    this->maxlat=maxlat;
    this->minlon=minlon;
    this->maxlon=maxlon;
    double midlat = (minlat+maxlat)/2.;
    mperlat = float(distance(minlat,0,maxlat,0)/(maxlat-minlat)*1000/1200);
    mperlon = float(distance(midlat,minlon,midlat,maxlon)/(maxlon-minlon)*1000/1200);
    latsteps = (maxlat-minlat)*resolution;
    lonsteps = (maxlon-minlon)*resolution*cos(midlat*0.0174532925);

    lineos = 3*pm->width() % 4;
    hc= new double*[lonsteps+2];
    dhdx= new double*[lonsteps+2];
    dhdy= new double*[lonsteps+2];
    //pxref = new px**[lonsteps+2];
    for(int wx = 0; wx < lonsteps+2; wx++){
        hc[wx] = new double[latsteps+2];
        dhdx[wx] = new double[latsteps+2];
        dhdy[wx] = new double[latsteps+2];
        //pxref[wx] = new px*[latsteps+2];
    }
    cd = new unsigned char[(3*pm->width()+lineos)*pm->height()];
}
void Terrain::set_terrain(double minlat,double maxlat,
                          double minlon, double maxlon){
    this->minlat=minlat;
    this->maxlat=maxlat;
    this->minlon=minlon;
    this->maxlon=maxlon;
    emaxlat.setText(QString("%1").arg(maxlat));
    eminlat.setText(QString("%1").arg(minlat));
    emaxlon.setText(QString("%1").arg(maxlon));
    eminlon.setText(QString("%1").arg(minlon));
    double midlat = (minlat+maxlat)/2.;
    mperlat = float(distance(minlat,0,maxlat,0)/(maxlat-minlat)*1000/1200);
    mperlon = float(distance(midlat,minlon,midlat,maxlon)/(maxlon-minlon)*1000/1200);
}

void Terrain::deletehdata(){
    for(int wx = 0; wx < lonsteps+2; wx++){
        delete hc[wx];
        delete dhdx[wx];
        delete dhdy[wx];
        //delete pxref;
    }
    delete hc;
    delete dhdx;
    delete dhdy;
    delete cd;
    //delete pxref;
}
void Terrain::add_DEM(){
    int x,y;
    int xo = pm->width() / 2;
    int yo = pm->height() / 2;
    for(int l = 0; l < latsteps; l++){
        double lat = minlat + (maxlat-minlat)*l/(latsteps-1);
        for(int k = 0; k < lonsteps; k++){
            double lon = minlon + (maxlon-minlon)*k/(lonsteps-1);
                //project(lat,lon,hc[k][l],Qt::black,&x,&y);
                for(int dx = -1; dx<2;dx++)
                    for(int dy = -1; dy < 2; dy++){
                    int xp =x+dx+xo;
                    int yp =y+dy+yo;
                    if(xp<pm->width() && xp>0 && yp>0 && yp<pm->height())
                        htocolorfast(&(cd[yp*(3*pm->width()+lineos)+xp*3]),
                                     hc[k][l],
                                     dhdx[k][l]/mperlon,
                                     dhdy[k][l]/mperlat);
                    }
        }
    }
    QImage img(cd,pm->width(),pm->height(),QImage::Format_RGB888);
    pm->convertFromImage(img);
}
inline void Terrain::htocolorfast(unsigned char* cd,float h,float dxf,float dyf){
    int i = 0;
    MapSettings* MS = &(settings->MS);
    while(h > MS->heights[i+1] && i < 3)
        i++;
    float l = qMax(0.,
              qMin(1.,
                   (MS->heights[i+1]-h)/(MS->heights[i+1]-MS->heights[i])));
    float r = 1-l;

    float I =  MS->sun.I*((-(dxf*MS->sun.sphi+dyf*MS->sun.cphi)*MS->sun.stheta
                           +MS->sun.ctheta)
               /sqrt(1+dxf*dxf+dyf*dyf)-MS->sun.ctheta);
    if(I>0){
        cd[0] = qMin(uchar(255),
                uchar((qRed  (MS->heightcolor[i])*l+qRed  (MS->heightcolor[i+1])*r)/(I+1)
                + I/(I+1)*255.));
        cd[1] = qMin(uchar(255),
                uchar((qGreen(MS->heightcolor[i])*l+qGreen(MS->heightcolor[i+1])*r)/(I+1)
                + I/(I+1)*255.));
        cd[2] = qMin(uchar(255),
                uchar((qBlue (MS->heightcolor[i])*l+qBlue (MS->heightcolor[i+1])*r)/(I+1)
                + I/(I+1)*255.));
    }
    else{
        cd[0] = uchar((qRed  (MS->heightcolor[i])*l+qRed  (MS->heightcolor[i+1])*r)/(-I+1));
        cd[1] = uchar((qGreen(MS->heightcolor[i])*l+qGreen(MS->heightcolor[i+1])*r)/(-I+1));
        cd[2] = uchar((qBlue (MS->heightcolor[i])*l+qBlue (MS->heightcolor[i+1])*r)/(-I+1));
    }
}
QColor Terrain::htocolor(int lati, int loni){
    unsigned char c[3];
    htocolorfast(c,hc[loni][lati],dhdx[loni][lati]/mperlon,dhdy[loni][lati]/mperlat);
    return QColor(c[0],c[1],c[2]);
}
QColor Terrain::htocolor(double h, double dlon, double dlat){
    unsigned char c[3];
    htocolorfast(c,h,dlon/mperlon,dlat/mperlat);
    return QColor(c[0],c[1],c[2]);
}
void Terrain::save(QString filename){
    pm->save(filename,"PNG");
}
void Terrain::mousePressEvent(QMouseEvent *event){
    int x = event->x()-Pic.x();
    int y = event->y()-Pic.x();
    if(x >= 0 && x<pm->width() && y >= 0 && y < pm->height()){
        if(event->button()==Qt::LeftButton)
            setpos(info[x][y].lat,info[x][y].lon,info[x][y].h+20,
                   azimutLE.text().toDouble(),elevLE.text().toDouble(),
                   zoomLE.text().toDouble());
        if(event->button()==Qt::RightButton)
            setpos(LatPos.text().toDouble(),
                   LonPos.text().toDouble(),
                   hPos.text().toDouble(),
                   info[x][y].lat,
                   info[x][y].lon,
                   info[x][y].h,
                   zoom);
    }
}
