#ifndef TERRAIN_H
#define TERRAIN_H

#include <QList>
#include "MAP.h"
#include "MathExt.h"
#include "terrainhelper.h"
#include <QPixmap>
#include <QPainter>
#include <QtWidgets>
#include <QtConcurrentRun>

typedef struct threadcontainer{
    MAP** m;
    HeightData *h;
    QImage *Img;
    QPainter *q;
    int xstart,xend;
}threadcontainer;

class Terrain : public QWidget{
    Q_OBJECT
public:
    Terrain();
    Terrain(OSM* o, Track** ct,int width, int height,config* c);
    ~Terrain();
    void setpos(double lat, double lon, double h,
                double azimut, double elev,double zoom);
    void setpos(double lat1, double lon1, double h1,
                double lat2, double lon2, double h2, double zoom);
    void set_terrain(double minlat,double maxlat,double minlon, double maxlon);
    void drawterrain(double minlat,double maxlat, double minlon, double maxlon,
                     double resolution);
    void terrain_raytrace(double minlat,double maxlat,
                          double minlon, double maxlon);
    void summit_collection(double lat, double lon, double h,QString filename,
                           QString label, double azimut0);
    void save(QString filename);
    void raytrace4(QImage *Img, MAP **m,QPainter *q,int xstart, int xend);
    QPixmap* getpixmap();
    QSize sizeHint() const;
    QSize minimumSizeHint() const;
    void paintEvent(QPaintEvent *);
    void mousePressEvent(QMouseEvent *event);
public slots:
    void recalc();
    void retexture();
    void save_png();
private:
    //px project(double lat, double lon, double h, QColor c,
    //           int*x, int*y);
    double elev;
    double zoom;                        //!< px/tan(angle of incidence);
    xyz R0,nv,nh,nz;
    QPixmap* pm;
    HeightData* h;
    MAP *m;
    double** hc;						/*!< height at px position*/
    double** dhdx;						/*!< gradient in x direction*/
    double** dhdy;						/*!< gradient in y direction*/
    pxinfo** info;                      /*!< gradient in y direction*/
    unsigned char* cd;					/*!< color data, every px consists of
                                        three bytes, used by all rendering*/
    double mperlon,mperlat;
    double minlat, maxlat,minlon, maxlon;
    void newhdata(double minlat,double maxlat,
                  double minlon, double maxlon,
                  double resolution);
    void deletehdata();
    void deletemaps();
    void raytrace(QImage* Img);
    void raytrace4(QImage *Img, MAP **m);
    void add_DEM();
    void htocolorfast(unsigned char* cd,float h,float dxf,float dyf);
    QColor htocolor(int lati, int loni);
    QColor htocolor(double h, double dlon, double dlat);
    int lineos,lonsteps,latsteps;
    config* settings;
    Track** t;
    QMutex qmutex;
    QImage Img[4];
    MAP* maps[4];
    //Widget elements
    QLabel Pic;
    QLineEdit LatPos, LonPos, hPos, azimutLE,elevLE,zoomLE,widthLE,heightLE,superele,fog,fogos,textureMpx,
    eminlat,emaxlat,eminlon,emaxlon;
    QLabel LLatPos, LLonPos, LhPos, LazimutLE,LelevLE,LzoomLE,LwidthLE,LheightLE,Lsuperele,Lfog,Lfogos,LtextureMpx,
    Leminlat,Lemaxlat,Leminlon,Lemaxlon;
    QPushButton BtnRecalc,BtnRetexture,Btnsave;
};

#endif // TERRAIN_H
