#include "tspmatrix.h"

TspMatrix::TspMatrix(double** data, int c){
    this->c = c;
    sum = 0;

    d = new double*[c];
    for(int i = 0; i < c; i++){
        d[i] = new double[c];
        for(int j = 0; j < c; j++){
            d[i][j] = data[i][j];
        }
        row.append(i);
        col.append(i);
    }
}
TspMatrix::~TspMatrix(){
    for(int i = 0; i < c; i++)
        delete d[i];
    delete d;
}
bool TspMatrix::finished(int maxi){
    return c == maxi;
}
void TspMatrix::primitive_solution(){
    int from = c-1;
    double *colc = new double[c];
    for(int i = 0; i < c; i++) colc[i] = d[i][c-1];
    for(int k = 0; k < c; k++){
        if(k == c-1) for(int j = 0; j < c; j++) d[j][c-1] = colc[j];
        double min = row_min(from);
        int i = 0;
        while (d[from][i] != min) i++;
        sum += min;
        edge* cdt = new edge;
        cdt->from = from;
        cdt->to = i;
        cdt->row = from;
        cdt->col = i;
        for(int j = 0; j < c; j++) d[j][from] = 1e9;
        from = i;
        solution.append(cdt);
    }
    // relink elements in solution
    for(int i = 0; i < solution.length(); i++){
        for(int j = i; j < solution.length(); j++){
            solution[i]->link_pre(solution[j]);
            solution[i]->link_suc(solution[j]);
        }
    }
    return;
}
edge* TspMatrix::get_solution(){
    return solution[0];
}

TspMatrix* TspMatrix::clone(){
    TspMatrix* T = new TspMatrix(d,c);
    T->fill_lists(&solution,&row,&col,sum);
    return T;
}
void TspMatrix::fill_lists(QList<edge*> *solution0, QList<int>* row0,
                           QList<int>* col0, double sum){
    row.clear();
    col.clear();
    solution.clear();
    for(int i = 0; i < row0->length(); i++)
        row.append(row0->at(i));
    for(int i = 0; i < col0->length(); i++)
        col.append(col0->at(i));
    for(int i = 0; i < solution0->length(); i++){
        solution.append(new edge(solution0->at(i)));
        solution[i]->reset_link();
    }
    // relink elements in solution
    for(int i = 0; i < solution.length(); i++){
        for(int j = i; j < solution.length(); j++){
            solution[i]->link_pre(solution[j]);
            solution[i]->link_suc(solution[j]);
        }
    }
    this->sum = sum;
}
QList<edge*>* TspMatrix::solutionlist(){
    return &solution;
}

double TspMatrix::lower_bound(){
    return sum;
}
void TspMatrix::deactivate_cell(int irow, int icolumn){
    d[irow][icolumn] = 1e9;
}
void TspMatrix::deactivate_cell(edge* e){
    deactivate_cell(e->row,e->col);
}

void TspMatrix::shrink(int irow, int icolumn){
    // remove row
    row.removeAt(irow);
    delete d[irow];
    for(int i = irow; i < c-1; i++)
        d[i] = d[i+1];
    // remove column
    col.removeAt(icolumn);
    for(int i = 0; i < c-1; i++)
        for(int j = icolumn; j < c-1; j++)
            d[i][j] = d[i][j+1];
    // reduce index
    c--;
}

void TspMatrix::reduce(){
    // first rows
    for(int i = 0; i < c; i++){
        double min = row_min(i);
        //double datacheck[6][6]; for(int k = 0; k < c; k++) for(int l = 0; l < c; l++)datacheck[k][l] = d[k][l];
        if(min > 0){
            sum += min;
            for(int j = 0; j < c; j++)
                d[i][j] -= min;
        }
    }
    // then columns
    for(int i = 0; i < c; i++){
        double min = col_min(i);
        if(min > 0){
            sum += min;
            for(int j = 0; j < c; j++)
                d[j][i] -= min;
        }
    }
}
double TspMatrix::row_min(int irow){
    double result = d[irow][0];
    for(int i = 1; i < c; i++)
        result = qMin(result, d[irow][i]);
    return result;
}
double TspMatrix::col_min(int icol){
    double result = d[0][icol];
    for(int i = 1; i < c; i++)
        result = qMin(result, d[i][icol]);
    return result;
}
double TspMatrix::row_min(int irow, int exc){
    double result;
    if(exc == 1)
        result = d[irow][0];
    else
        result = d[irow][1];

    for(int i = 0; i < c; i++)
        if(i != exc && d[irow][i] < result)
        result = d[irow][i];

    return result;
}

double TspMatrix::col_min(int icol, int exc){
    double result;
    if(exc == 1)
        result = d[0][icol];
    else
        result = d[1][icol];

    for(int i = 0; i < c; i++)
        if(i != exc && d[i][icol] < result)
        result = d[i][icol];

    return result;
}

edge* TspMatrix::candidate(double* penalty){
    edge* cdt = new edge;
    cdt->from = -1;
    double cval = 1e9;
    *penalty = -1;
    // get total minimum
    for(int k = 0; k < c; k++)
        cval = qMin(cval,row_min(k));
    // investigate the candidates
    for(int i = 0; i < c; i++)
        for(int j = 0; j < c; j++)
            if(d[i][j] == cval){
                double pt = col_min(j,i) + row_min(i,j);
                if(pt > *penalty){
                    cdt->from = row[i];
                    cdt->to = col[j];
                    cdt->row = i;
                    cdt->col = j;
                    *penalty = pt;
                }
            }
    return cdt;
}
double TspMatrix::accept_candidate(){
    double penalty = 0;
    edge* cdt = candidate(&penalty);
    // check for subtour
    double linked = false;
    //double datacheck[6][6]; for(int k = 0; k < c; k++) for(int l = 0; l < c; l++)datacheck[k][l] = d[k][l];
    for(int i = 0; i < solution.length(); i++){
        if(cdt->link_pre(solution[i])){
            // try to avoid closure of subtour in next iteration
            int ir = row.indexOf(solution[i]->to0());
            int ic = col.indexOf(cdt->from0());
            if(c > 2 && ir >= 0 && ir < c && ic >= 0 && ic < c)
                //
                d[ir][ic] = 1e9;
            linked = true;
        }
        if(cdt->link_suc(solution[i])){
            // try to avoid closure of subtour in next iteration
            int ir = row.indexOf(cdt->to0());
            int ic = col.indexOf(solution[i]->from0());
            if(c > 2 && ir >= 0 && ir < c && ic >= 0 && ic < c)
                d[ir][ic] = 1e9;
            linked = true;
        }
        if(linked)
            i == solution.length()+1;
    }
    // minimum subtour avoidence
    //for(int k = 0; k < c; k++) for(int l = 0; l < c; l++)datacheck[k][l] = d[k][l];
    if(!linked){
        int ic = col.indexOf(cdt->from);
        int ir = row.indexOf(cdt->to);
        if(ir >= 0 && ir < c && ic >= 0 && ic < c)
            d[ir][ic] = 1e9;
    }
    // know add to solution and remove from matrix
    solution.append(cdt);
    shrink(cdt->row,cdt->col);
    //for(int k = 0; k < c; k++) for(int l = 0; l < c; l++)datacheck[k][l] = d[k][l];
    return sum;
}

double TspMatrix::refuse_candidate(){
    if(c == 1)
        return 1e9;
    double penalty = 0;
    edge* cdt = candidate(&penalty);
    deactivate_cell(cdt);
    delete cdt;
    return sum + penalty;
}
edge::edge(){prev = NULL; next = NULL;}
edge::edge(edge* src){
    prev = NULL; next = NULL;
    from = src->from;
    to = src->to;
    row = col = -1;
}

bool edge::is_precessor_of(edge *l){
    return to == l->from;
}
bool edge::is_succecessor_of(edge *l){
    return from == l->to;
}
int edge::from0(){
    edge *t = this;
    while(t->prev != NULL && t->prev != this)
        t = t->prev;
    return t->from;
}

int edge::to0(){
    edge *t = this;
    while(t->next != NULL && t->next != this)
        t = t->next;
    return t->to;

}
bool edge::link_pre(edge* l){
    if(l->prev == NULL){
        if(l->from == to){
            l->prev = this;
            this->next = l;
            return true;
        }
        else
            return false;
    }
    else
        return false;
}

bool edge::link_suc(edge* l){
    if(l->next == NULL){
        if(l->to == from){
            l->next = this;
            this->prev = l;
            return true;
        }
        else
            return false;
    }
    else
        return false;

}
void edge::reset_link(){
    prev = next = NULL;
}
edge* edge::start(){
    edge* result = this, *t = result, *t0 = result;
    do{ result = t;
        t = t->bw();
    }while(t != NULL && t != t0);
    return result;
}

edge* edge::fw(){
    return next;
}

edge* edge::bw(){
    return prev;
}
