#include "webdata.h"


WebData::WebData(webdatacmd w){
    this->w = w;
}

void WebData::run(){
    switch(w.mode){
        case CreateHTML:
            analyse_climbs();
            break;
        case ReadOSM:
            read_osm();
            break;
        case CreateTiles:
            generate_map();
            break;
        case CreateTilesBTP:
            generate_map_BTP();
            break;
        case SplitEurope:
            split_europe();
            break;
        case CreateSTRONGLayers:
            generate_STRONG();
            break;
    }
}

void WebData::exit(){}
void WebData::generate_map(){
    QDir maind(w.dest);
    QDir zdir(QString(w.dest).append(QString("\\%1").arg(w.z)));
    QDir ydir(QString(w.dest).append(QString("\\%1\\%2").arg(w.z).arg(w.y)));
    if(!zdir.exists()) maind.mkdir(zdir.dirName());
    if(!ydir.exists()) zdir.mkdir(ydir.dirName());
    HeightData hd;
    hd.set_SRTM3_folder(w.c->SRTM3folder);
    MAP mt(&hd,w.osm);
    mt.settings(&(w.c->MS));
    mt.resize(w.width,w.height);
    mt.set_borders(yztolat(w.y+1,w.z),yztolat(w.y,w.z),xztolon(w.x,w.z),xztolon(w.x+1,w.z));
    mt.calc_borders();
    mt.draw_map(NULL,NULL,NULL,0);
    mt.get_Map()->save(ydir.absoluteFilePath(QString("%1.png").arg(w.x)),"PNG");
}
void WebData::split_europe(){
    double dd = 0.15;
    QString filename = QString("D:\\ClimbProject\\pbf-Data\\%1-%2-%3.pbf").arg(w.z).arg(w.x).arg(w.y);
    QFileInfo fi = QFileInfo(filename);
    qint64 fs = fi.size();
    if(fi.exists()){
        if((w.z < 9 || fi.size() > 400e6) && fs > 250)
            for(int jx = 0; jx < 2; jx++)
                for(int jy = 0; jy < 2; jy++){
                    webdatacmd* wt = new webdatacmd;
                    wt->z = w.z + 1;
                    wt->x = w.x*2 + jx;
                    wt->y = w.y*2 + jy;
                    wt->mode = SplitEurope;
                    double minlat = qMax( -90.,yztolat(wt->y+1,wt->z) - dd);
                    double maxlat = qMin(  90.,yztolat(wt->y,wt->z)   + dd);
                    double minlon = qMax(-180.,xztolon(wt->x,wt->z)   - dd);
                    double maxlon = qMin( 180.,xztolon(wt->x+1,wt->z) + dd);
                    QString newfile = QString("D:\\ClimbProject\\pbf-Data\\%1-%2-%3.pbf").arg(wt->z).arg(wt->x).arg(wt->y);
                    QString cmdline = QString("D:\\ClimbProject\\pbf-Data\\osmconvert.exe");
                    QStringList arguments;
                    arguments << filename;
                    arguments << QString("-b=%1,%2,%3,%4").arg(minlon)
                                                         .arg(minlat)
                                                         .arg(maxlon)
                                                         .arg(maxlat);
                    arguments << QString("-o=").append(newfile);
                    //QString cmdline = QString("C:\\Users\\User\c++\\BTP3\\Release\\BTP3.exe");
                    //QStringList arguments;
                    //arguments << "close";
                    QProcess System_Call;
                    System_Call.start(cmdline,arguments);
                    System_Call.waitForFinished(-1);
                    emit add(wt);
                }
    }
}
double WebData::xztolon(int x, int z){
    double step = 360. / pow(2.,z);
    return x*step - 180.;
}
double WebData::yztolat(int y, int z){
    double step = 360. / pow(2.,z);
    return mercatortolat(180-y*step);
}
void WebData::read_osm(){
    QFileInfo fi = QFileInfo(w.src);
    qint64 fs = fi.size();
    if(fi.exists() && fs > 250){
        QString osmout(w.dest);
        osmout.append(".osm");
        QString cmdline = QString("D:\\ClimbProject\\pbf-Data\\osmconvert.exe");
        QStringList arguments;
        arguments << w.src;
        arguments << QString("-o=").append(osmout);
        QProcess System_Call;
        System_Call.start(cmdline,arguments);
        System_Call.waitForFinished(-1);
        cmdline = QString("C:\\Users\\User\\c++\\BTP3\\release\\BTP3.exe");
        QStringList argumentBTP;
        argumentBTP << QString("-r=").append(osmout);
        argumentBTP << QString("-d=").append(w.dest);
        argumentBTP << QString("close");
        System_Call.start(cmdline,argumentBTP);
        System_Call.waitForFinished(-1);
        QFile(osmout).remove();
    }
}
void WebData::analyse_climbs(){
    QDir di(w.dest);
    QString dirname = di.dirName();
    if(!di.exists()){
        // create directory as it does not exist
        di.cdUp();
        di.mkdir(dirname);
    }
    QString srcfilename = QString(w.dest).append("\\")
                                         .append(dirname)
                                         .append("-BinaryList.src");
    FILE* srcfile = fopen(srcfilename.toLocal8Bit().data(),"wb");
    int zero = 0;
    fwrite(&zero,4,1,srcfile);
    fclose(srcfile);
            srcfilename = QString(w.dest).append("\\")
                                         .append(dirname)
                                         .append("-names.txt");
    srcfile = fopen(srcfilename.toLocal8Bit().data(),"w");
    fclose(srcfile);
    QFileInfo fi = QFileInfo(w.src);
    qint64 fs = fi.size();
    if(fi.exists() && fs > 108){
        QString dummyfile = QString(w.dest).append(".dmy");
        FILE* dmy = fopen(dummyfile.toLocal8Bit().data(),"w");
        fclose(dmy);
        double minlat = qMax( -90.,yztolat(w.y+1,w.z));
        double maxlat = qMin(  90.,yztolat(w.y,w.z)  );
        double minlon = qMax(-180.,xztolon(w.x,w.z)  );
        double maxlon = qMin( 180.,xztolon(w.x+1,w.z));
        QString cmdline = QString("C:\\Users\\User\\c++\\BTP3\\release\\BTP3.exe");
        QStringList arguments;
        arguments << w.src;
        arguments << QString("-b=%1,%2,%3,%4").arg(minlon)
                                             .arg(minlat)
                                             .arg(maxlon)
                                             .arg(maxlat);
        arguments << QString("-a=").append(w.dest);
        arguments << QString("close");
        QProcess System_Call;
        System_Call.start(cmdline,arguments);
        System_Call.waitForFinished(-1);
        QFile fi2(dummyfile);
        if(fi2.exists()) fi2.remove();
    }
}
void WebData::generate_map_BTP(){
    QDir maind(w.dest);
    HeightData hd;
    hd.set_SRTM3_folder(w.c->SRTM3folder);
    OSM osm;
    for(int dx = -1; dx < 2; dx++)
        for(int dy = -1; dy < 2; dy++)
            osm.load(QString("%1%4-%2-%3.btp").arg(w.src).arg(w.x+dx)
                                              .arg(w.y+dy).arg(9)
                                              .toLocal8Bit().data());

    MAP mt(&hd,&osm);
    mt.settings(&(w.c->MS));

    //first zoomlevel 9 .. 11
    for(int z = 9; z < 12; z++){
        QDir zdir(QString(w.dest).append(QString("\\%1").arg(z)));
        if(!zdir.exists()) maind.mkdir(zdir.dirName());
        // generate the map
        int m = pow(2,z-w.z);
        mt.resize(w.width*3*m,w.height*3*m);
        mt.set_borders(yztolat(w.y+2,w.z),yztolat(w.y-1,w.z),
                       xztolon(w.x-1,w.z),xztolon(w.x+2,w.z));
        mt.calc_borders();
        mt.draw_map(NULL,NULL,NULL,1);
        // extract all tiles from it
        for(int y = 0; y < m; y++){
            QDir ydir(QString(w.dest).append(QString("\\%1\\%2").arg(z).arg(w.y*m+y)));
            if(!ydir.exists()) zdir.mkdir(ydir.dirName());
            for(int x = 0; x < m; x++){
                QPixmap pt = mt.get_Map()->copy((m+x)*256,(m+y)*256,256,256);
                QString filename = QString("%1%2\\%3\\%4.png").arg(w.dest)
                                                            .arg(z)
                                                            .arg(w.y*m+y)
                                                            .arg(w.x*m+x);
                pt.save(filename,"PNG");
            }
        }
    }
    //then zoomlevel 12 .. 14
    webdatacmd wt = w;
    wt.z = 12;
    for(wt.x = w.x*8; wt.x < (w.x+1)*8; wt.x++)
        for(wt.y = w.y*8; wt.y < (w.y+1)*8; wt.y++)
            for(int z = 12; z < 15; z++){
                QDir zdir(QString(w.dest).append(QString("\\%1").arg(z)));
                if(!zdir.exists()) maind.mkdir(zdir.dirName());
                // generate the map
                int m = pow(2,z-wt.z);
                mt.resize(wt.width*3*m,wt.height*3*m);
                mt.set_borders(yztolat(wt.y+2,wt.z),yztolat(wt.y-1,wt.z),
                               xztolon(wt.x-1,wt.z),xztolon(wt.x+2,wt.z));
                mt.calc_borders();
                mt.draw_map(NULL,NULL,NULL,1);
                // extract all tiles from it
                for(int y = 0; y < m; y++){
                    QDir ydir(QString(w.dest).append(QString("\\%1\\%2").arg(z).arg(wt.y*m+y)));
                    if(!ydir.exists()) zdir.mkdir(ydir.dirName());
                    for(int x = 0; x < m; x++){
                        QPixmap pt = mt.get_Map()->copy((m+x)*256,(m+y)*256,256,256);
                        QString filename = QString("%1%2\\%3\\%4.png").arg(wt.dest)
                                                                    .arg(z)
                                                                    .arg(wt.y*m+y)
                                                                    .arg(wt.x*m+x);
                        pt.save(filename,"PNG");
                    }
                }
            }
}
void WebData::generate_STRONG(){
    QDir maind(w.dest);
    HeightData hd;
    hd.set_SRTM3_folder(w.c->SRTM3folder);
    OSM osm;
    osm.load(QString("%1%4-%2-%3.btp").arg(w.src).arg(w.x).arg(w.y).arg(9).toLocal8Bit().data());
    //osm.normalize_Hm(osm.get_first_KOO());
    osm.invert_Hm(osm.get_first_KOO());
    int N;
    Track** t;
    //STRONGcalc s(&osm,NULL,NULL,w.c->maxtype,w.c->mintype,&(w.c->sconfig),&t,&N,true);
    STRONGcalc s(&osm,NULL,NULL,w.c->maxtype,w.c->mintype,&(w.c->cconfig),&t,&N,false);
    s.start();
    s.wait();
    int Nmax = 0;
    for(int i = 0; i < N; i++)
        if(t[i] != NULL){
            Nmax = i;
            delete t[i];
        }
    delete t;
    Routing r(&osm);
    if(Nmax > 0){
        STRONGlayer* SL = r.STRONGlayer_create(Nmax);
        QString SLfile = QString("%1%4-%2-%3.bsl").arg(w.dest).arg(w.y).arg(w.x).arg(9);
        r.STRONGlayer_save(SLfile.toLocal8Bit().data(),SL);

        MAP mt(&hd,&osm);
        mt.settings(&(w.c->MS));

        //first zoomlevel 9 .. 11
        for(int z = 9; z < 12; z++){
            QDir zdir(QString(w.dest).append(QString("\\%1").arg(z)));
            if(!zdir.exists()) maind.mkdir(zdir.dirName());
            // generate the map
            int m = pow(2,z-w.z);
            mt.resize(w.width*3*m,w.height*3*m);
            mt.set_borders(yztolat(w.y+2,w.z),yztolat(w.y-1,w.z),
                           xztolon(w.x-1,w.z),xztolon(w.x+2,w.z));
            mt.calc_borders();
            QImage image(mt.getwidth(),mt.getheight(),QImage::Format_ARGB32);
            image.fill(qRgba(0,0,0,0));
            QPainter p(&image);
            p.setRenderHint(QPainter::Antialiasing);
            mt.STRONGlayer_to_Map(&p,SL);
            // extract all tiles from it
            for(int y = 0; y < m; y++){
                QDir ydir(QString(w.dest).append(QString("\\%1\\%2").arg(z).arg(w.y*m+y)));
                if(!ydir.exists()) zdir.mkdir(ydir.dirName());
                for(int x = 0; x < m; x++){
                    QImage it = image.copy((m+x)*256,(m+y)*256,256,256);
                    QString filename = QString("%1%2\\%3\\%4.png").arg(w.dest)
                                                                .arg(z)
                                                                .arg(w.y*m+y)
                                                                .arg(w.x*m+x);
                    it.save(filename,"PNG");
                }
            }
        }
        //then zoomlevel 12 .. 14
        webdatacmd wt = w;
        wt.z = 12;
        for(wt.x = w.x*8; wt.x < (w.x+1)*8; wt.x++)
            for(wt.y = w.y*8; wt.y < (w.y+1)*8; wt.y++)
                for(int z = 12; z < 15; z++){
                    QDir zdir(QString(w.dest).append(QString("\\%1").arg(z)));
                    if(!zdir.exists()) maind.mkdir(zdir.dirName());
                    // generate the map
                    int m = pow(2,z-wt.z);
                    mt.resize(wt.width*3*m,wt.height*3*m);
                    mt.set_borders(yztolat(wt.y+2,wt.z),yztolat(wt.y-1,wt.z),
                                   xztolon(wt.x-1,wt.z),xztolon(wt.x+2,wt.z));
                    mt.calc_borders();
                    QImage image(mt.getwidth(),mt.getheight(),QImage::Format_ARGB32);
                    image.fill(qRgba(0,0,0,0));
                    QPainter p(&image);
                    p.setRenderHint(QPainter::Antialiasing);
                    mt.STRONGlayer_to_Map(&p,SL);
                    // extract all tiles from it
                    for(int y = 0; y < m; y++){
                        QDir ydir(QString(w.dest).append(QString("\\%1\\%2").arg(z).arg(wt.y*m+y)));
                        if(!ydir.exists()) zdir.mkdir(ydir.dirName());
                        for(int x = 0; x < m; x++){
                            QImage it = image.copy((m+x)*256,(m+y)*256,256,256);
                            QString filename = QString("%1%2\\%3\\%4.png").arg(wt.dest)
                                                                        .arg(z)
                                                                        .arg(wt.y*m+y)
                                                                        .arg(wt.x*m+x);
                            it.save(filename,"PNG");
                        }
                    }
                }
        delete SL->data;
        delete SL;
    }
}
